import inspect
import sheraf
from sheraf.models.indexation import model_from_table
from sheraf.attributes import Attribute
from sheraf.attributes.collections import ListAttribute, SetAttribute


class ModelLoader(object):
    """
    Loads models from the base in a cache.

    Inherited by most model types
    (``Model[Attribute|List|Set|(Large)Dict]``)
    """

    cache = {}

    def __init__(self, model=None, **kwargs):
        super().__init__(**kwargs)
        self._model = model
        frame = inspect.stack()[3]
        module = inspect.getmodule(frame[0])
        self.module_path = module.__name__ if module else None

    def load_model(self, modelpath):
        # Internal.
        # :param modelpath: namespace of the model to load (eg 'x.y.z')
        # :return: a model class given its path
        if isinstance(modelpath, bytes):
            modelpath = modelpath.decode("utf-8")

        path = modelpath.split(".")
        module_path, klass = ".".join(path[:-1]), path[-1]
        module_path = module_path or self.module_path

        module = __import__(module_path, globals(), locals(), [klass], 0)

        # pypy:
        if module is None:  # pragma: no cover
            raise ImportError(modelpath)

        try:
            return getattr(module, klass)
        except AttributeError as exc:
            raise ImportError(exc)

    def read(self, parent):
        self.check_model(parent)
        return super().read(parent)

    def write(self, parent, value):
        self.check_model(parent)
        return super().write(parent, value)

    @property
    def model(self):
        self.check_model()
        return self._model

    def check_model(self, parent=None):
        if isinstance(self._model, (list, tuple)):
            self._model = type(self._model)(
                self._check_model(m, parent) for m in self._model
            )
        else:
            self._model = self._check_model(self._model, parent)

    def _check_model(self, model, parent):
        if isinstance(model, (str, bytes)):
            try:
                return ModelLoader.cache[model]
            except KeyError:
                ModelLoader.cache[model] = self.load_model(model)
                return ModelLoader.cache[model]

        elif parent and not isinstance(model, type):
            return type(
                "{}.{}".format(parent.__class__.__name__, self.key(parent)),
                (model.__class__,),
                model.attributes,
            )

        else:
            return model


class AttributeLoader(ModelLoader):
    def __init__(self, attribute=None, **kwargs):
        self._attribute = attribute
        super().__init__(**kwargs)

    def check_attribute(self, attribute_name, parent):
        if attribute_name not in self._model.attributes:
            raise sheraf.SherafException(
                f"'{attribute_name}' is not an attribute of {self._model.__name__}"
            )

        attribute = self._model.attributes[attribute_name]
        if isinstance(attribute, (ListAttribute, SetAttribute)):
            if not isinstance(attribute.attribute, ModelAttribute):
                raise sheraf.SherafException(
                    f"'{self._model.__name__}.{attribute}' attribute should hold a 'ModelAttribute' to be referenced by a 'ReverseModelAttribute'"
                )

        elif not isinstance(attribute, ModelAttribute):
            raise sheraf.SherafException(
                f"'{self._model.__name__}.{attribute}' should be a 'ModelAttribute' or a collection of 'ModelAttribute' to be referenced by a 'ReversedModelAttribute'"
            )

        if attribute_name not in self.model.indexes:
            raise sheraf.SherafException(
                f"'{self._model.__name__}.{attribute}' should have an index to be referenced by a 'ReversedModelAttribute'"
            )


class ModelAttribute(ModelLoader, Attribute):
    """This attribute references another :class:`~sheraf.models.Model`.

    :param model: The model type to store.
    :type model: :class:`~sheraf.models.Model` or list of :class:`~sheraf.models.Model`

    >>> class Horse(sheraf.Model):
    ...     table = "horse"
    ...     name = sheraf.SimpleAttribute()
    ...
    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.SimpleAttribute()
    ...     mount = sheraf.ModelAttribute(Horse)
    ...
    >>> with sheraf.connection(commit=True):
    ...     jolly = Horse.create(name="Jolly Jumper")
    ...     george = Cowboy.create(name="George Abitbol", mount=jolly)
    ...
    ...     george.mount.name
    'Jolly Jumper'

    The referenced model can be dynamically created if its structure is passed through as a dict:

    >>> with sheraf.connection(commit=True):
    ...     peter = Cowboy.create(name="Peter", mount={"name": "Polly Pumper"})
    ...     assert isinstance(peter.mount, Horse)
    ...     peter.mount.name
    'Polly Pumper'

    When the referenced model is deleted, the value of the attribute becomes ``None``.

    >>> with sheraf.connection(commit=True):
    ...     george = Cowboy.read(george.id)
    ...     jolly.delete()
    ...     assert george.mount is None

    Several model classes can be used, but this will be more memory consuming in the database.

    >>> class Pony(sheraf.Model):
    ...     table = "pony"
    ...     name = sheraf.SimpleAttribute()
    ...
    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.SimpleAttribute()
    ...     mount = sheraf.ModelAttribute((Horse, Pony))
    ...
    >>> with sheraf.connection(commit=True):
    ...     superpony = Pony.create(name="Superpony")
    ...     peter = Cowboy.create(name="Peter", mount=superpony)

    When several models are set, the first one is considered to be the default model.
    The default model is used when there is a doubt on the read data, or in the case
    of model creation with a dict.
    """

    def __init__(self, model=None, **kwargs):
        if not model:
            raise sheraf.exceptions.SherafException(
                "ModelAttribute requires model parameter."
            )
        kwargs["read_memoization"] = False
        kwargs["write_memoization"] = False
        super().__init__(default=None, model=model, **kwargs)

    def values(self, model):
        """
        By default :class:`~sheraf.attributes.models.ModelAttribute` are indexed on
        their identifier.
        """
        if model is None:
            return {None}

        return {
            (model.table, model.identifier)
            if isinstance(self.model, (tuple, list))
            else model.identifier
        }

    def deserialize(self, value):
        if isinstance(value, tuple):
            table, id_ = value
            model = model_from_table(table)
        else:
            model = (
                self.model[0] if isinstance(self.model, (list, tuple)) else self.model
            )
            id_ = value

        try:
            return model.read(id_)
        except (KeyError, sheraf.exceptions.ModelObjectNotFoundException):
            return None

    def serialize(self, value):
        if value is None:
            return None

        elif isinstance(value, sheraf.IndexedModel):
            return (
                (value.table, value.identifier)
                if isinstance(self.model, (tuple, list))
                else value.identifier
            )

        elif self.model and isinstance(value, dict):
            return self.model.create(**value).identifier

        else:
            try:
                return self.model.read(value).identifier
            except sheraf.SherafException:
                return None

    def update(
        self,
        old_value,
        new_value,
        addition=True,
        edition=True,
        deletion=False,
        replacement=False,
    ):
        if replacement or old_value is None or not isinstance(new_value, dict):
            return self.serialize(new_value)

        return old_value.edit(new_value, addition, edition, deletion, replacement)


class ReverseModelAttribute(AttributeLoader, Attribute):
    """
    Inverse reference to a :class:`~sheraf.attributes.models.ModelAttribute`.

    :param model: The :class:`~sheraf.models.Model` to refer to.
    :param attribute: The :class:`~sheraf.attributes.Attribute` in the model to refer to.
        This model must be a :class:`~sheraf.attributes.models.ModelAttribute` or a
        collection of :class:`~sheraf.attributes.models.ModelAttribute`.

    The referenced attribute must be indexed.

    >>> class Cowboy(sheraf.Model):  # doctest: +SKIP
    ...     table = "reverse_cowboys"
    ...     name = sheraf.StringAttribute()
    ...     horse = sheraf.ModelAttribute("Horse").index()
    ...
    >>> class Horse(sheraf.Model):  # doctest: +SKIP
    ...     table = "reverse_horses"
    ...     name = sheraf.StringAttribute()
    ...     cowboy = sheraf.ReverseModelAttribute("Cowboy", "horse")
    ...
    >>> with sheraf.connection():  # doctest: +SKIP
    ...     george = Cowboy.create(name="George")
    ...     horse = Horse.create(name="Jolly", cowboy=george)
    ...     george.horse.name
    "Jolly"

    Collection attributes are also supported:

    >>> class Cowboy(sheraf.Model):  # doctest: +SKIP
    ...     table = "reverse_multicowboys"
    ...     name = sheraf.StringAttribute()
    ...     horses = sheraf.LargeListAttribute(ModelAttribute("Horse").index())
    ...
    >>> class Horse(sheraf.Model):  # doctest: +SKIP
    ...     table = "reverse_multihorses"
    ...     name = sheraf.StringAttribute()
    ...     cowboy = sheraf.ReverseModelAttribute("Cowboy", "horses")
    ...
    >>> with sheraf.connection():  # doctest: +SKIP
    ...     george = Cowboy.create(name="George")
    ...     jolly = Horse.create(name="Jolly", cowboy=george)
    ...     polly = Horse.create(name="Polly", cowboy=george)
    ...     george.horses[0].name
    ...     george.horses[1].name
    "Jolly"
    "Polly"
    """

    def __init__(self, model, attribute, **kwargs):
        kwargs["read_memoization"] = False
        kwargs["write_memoization"] = False
        super().__init__(default=None, model=model, attribute=attribute, **kwargs)

    def read(self, parent):
        self.check_model(parent)
        self.check_attribute(self._attribute, parent)
        search_args = {self._attribute: parent}

        if not self.model.indexes[self._attribute].details.unique:
            return self.model.search(**search_args)

        try:
            return self.model.search(**search_args).get()
        except sheraf.QuerySetUnpackException:
            return None

    def write(self, parent, value):
        self.check_model(parent)
        self.check_attribute(self._attribute, parent)

        self.delete(parent)

        if value is None:
            return None

        if not isinstance(value, (list, set)):
            value = [value]

        # if values are ids, then we should load corresponding models
        checked_values = []
        for v in value:
            if isinstance(v, self.model):
                checked_values.append(v)
            else:
                try:
                    checked_values.append(self.model.read(v))
                except sheraf.ModelObjectNotFoundException:
                    pass

        for referent in checked_values:
            attribute = self.model.attributes[self._attribute]

            if isinstance(attribute, ModelAttribute):
                setattr(referent, self._attribute, parent)

            if isinstance(attribute, sheraf.ListAttribute):
                setattr(
                    referent,
                    self._attribute,
                    getattr(referent, self._attribute) + [parent],
                )

            if isinstance(attribute, sheraf.SetAttribute):
                setattr(
                    referent,
                    self._attribute,
                    getattr(referent, self._attribute) | {parent},
                )

        return checked_values

    def delete(self, parent):
        referents = self.read(parent)

        if referents is None:
            return

        if isinstance(referents, self.model):
            referents = [referents]

        for referent in list(referents):
            attribute = self.model.attributes[self._attribute]

            if isinstance(attribute, ModelAttribute):
                delattr(referent, self._attribute)

            if isinstance(
                attribute,
                sheraf.ListAttribute,
            ):
                new_values = [
                    e for e in getattr(referent, self._attribute) if e != parent
                ]
                setattr(
                    referent,
                    self._attribute,
                    new_values,
                )

            if isinstance(
                attribute,
                sheraf.SetAttribute,
            ):
                new_values = {
                    e for e in getattr(referent, self._attribute) if e != parent
                }
                setattr(
                    referent,
                    self._attribute,
                    new_values,
                )


class InlineModelAttribute(ModelLoader, Attribute):
    """:class:`~sheraf.attributes.models.ModelAttribute` behaves like a basic
    model (i.e. have no indexation capability). The child attribute mapping is stored
    in the parent mapping.

    :param model: The model type to store.
    :type model: :class:`~sheraf.models.inline.InlineModel`

    >>> class Horse(sheraf.InlineModel):
    ...     name = sheraf.StringAttribute()
    ...
    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy_inliner"
    ...     name = sheraf.StringAttribute()
    ...     horse = sheraf.InlineModelAttribute(Horse)
    ...
    >>> with sheraf.connection(commit=True):
    ...     jolly = Horse.create(name="Jolly Jumper")
    ...     george = Cowboy.create(name="George", horse=jolly)
    ...     george.horse.name
    'Jolly Jumper'
    """

    def __init__(self, model=None, **kwargs):
        kwargs.setdefault("default", sheraf.types.SmallDict)
        super().__init__(model=model, **kwargs)

    def deserialize(self, value):
        if value is None:
            return None

        return self.model._decorate(value)

    def serialize(self, value):
        if value is None:
            return None

        elif isinstance(value, sheraf.InlineModel):
            return value.mapping

        elif isinstance(value, dict):
            return self.model.create(**value).mapping

        else:
            return self._default_value(value)

    def update(
        self,
        old_value,
        new_value,
        addition=True,
        edition=True,
        deletion=False,
        replacement=False,
    ):
        if replacement or old_value is None:
            return self.serialize(new_value)

        return old_value.edit(new_value, addition, edition, deletion, replacement)


class IndexedModelAttribute(ModelLoader, Attribute):
    """
    :class:`~sheraf.attributes.models.ModelAttribute` behaves like a classic model,
    including the indexation capabilities. The child attribute mapping and all the
    index mappings are is stored in the parent mapping.

    :param model: The model type to store.
    :type model: :class:`~sheraf.models.AttributeModel`

    .. note:: The :class:`~sheraf.models.AttributeModel` must have a *primary* index.

    >>> class Horse(sheraf.AttributeModel):
    ...     name = sheraf.StringAttribute().index(primary=True)
    ...     size = sheraf.IntegerAttribute().index()
    ...
    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy_indexer"
    ...     name = sheraf.StringAttribute()
    ...     horses = sheraf.IndexedModelAttribute(Horse)
    ...
    >>> with sheraf.connection(commit=True):
    ...     george = Cowboy.create(name="George Abitbol")
    ...     jolly = george.horses.create(name="Jolly Jumper", size=32)
    ...
    ...     assert jolly == george.horses.read("Jolly Jumper")
    ...     assert jolly in george.horses.search(size=32)
    """

    def read(self, parent):
        for index in self.model.indexes.values():
            key = self.key(parent)
            if key not in parent.mapping:
                parent.mapping[key] = sheraf.types.SmallDict()
            index.persistent = parent.mapping[key]

        return self.model

    def write(self, parent, value):
        model = self.read(parent)
        for values_dict in value:
            model.create(**values_dict)
        return model



from BTrees.OOBTree import OOBTree
from sheraf.attributes.index import Index


def set_read_memoization(should_memoize_read):
    Attribute.read_memoization = should_memoize_read


def set_write_memoization(should_memoize_write):
    Attribute.write_memoization = should_memoize_write


class Attribute(object):
    """
    Base type of all attributes of a base model.

    :param default: The value this attribute will be initialized with. If it is a callable object,
                    it will be called on initialization, else it will simply be copied. The callable
                    object can take either no argument, or one argument that will be the parent model.
    :type default: a callable object or a simple object
    :param key: The key to identify the attribute in its parent persistent mapping.
    :param lazy: If True, the objet carried by the attribute is created on the first
                          read or write access. If False, it is created when the model object
                          is created. Default is True.
    :type lazy: :class:`bool`
    :param read_memoization: Whether this attribute should be memoized on read. ``False`` by default.
    :type read_memoization: :class:`bool`
    :param write_memoization: Whether this attribute should be memoized on write. ``True`` by default.
    :type write_memoization: :class:`bool`

    When an attribute is memoized, its next reading will not result in a new database access.
    Attributes:
    - indexes:    a dictionary of Indexes. The key with value None stands for this attribute's name.
    """

    default_index_mapping = OOBTree
    nullok = True
    noneok = False
    read_memoization = False
    write_memoization = True

    def __init__(
        self,
        default=None,
        key=None,
        lazy=True,
        read_memoization=None,
        write_memoization=None,
        store_default_value=True,
    ):
        self._default_value = default
        self.attribute_name = None
        self._key = key

        if read_memoization is not None:
            self.read_memoization = read_memoization

        if write_memoization is not None:
            self.write_memoization = write_memoization

        self.lazy = lazy
        self.store_default_value = store_default_value
        self.indexes = {}
        self.cb_creation = []
        self.cb_edition = []
        self.cb_deletion = []

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.attribute_name}>"

    def create(self, parent):
        """
        :return: The attribute's defined default value. If it is a callable, return
                 the result of calling it instead.
        """
        if not callable(self._default_value):
            return self._default_value

        try:
            return self._default_value()
        except TypeError:
            return self._default_value(parent)

    def is_created(self, parent):
        """
        :param parent: The owner of this attribute
        :return: true if the object is effectively created"""
        return self.key(parent) in parent.mapping

    def key(self, parent):
        # the key that identifies this attribute in its owner object
        if self._key is None:
            return self.attribute_name

        if isinstance(self._key, (list, tuple)):
            for key in self._key:
                if key in parent.mapping:
                    return key
            return self._key[0]

        return self._key

    def read_raw(self, parent):
        # Internal.
        # :param parent: The owner of this attribute
        # :return: the raw representation of this attribute (ie as stored in ZODB)
        try:
            return parent.mapping[self.key(parent)]
        except KeyError:
            default_value = self.serialize(self.create(parent))
            if self.store_default_value:
                parent.mapping[self.key(parent)] = default_value
            return default_value

    def write_raw(self, parent, value):
        # Internal.
        # :param parent: The owner of this attribute
        # :param value: The value assigned to this attribute
        # assigns the value parameter to this attribute
        parent.mapping[self.key(parent)] = value
        return value

    def serialize(self, value):
        # Get data and transform it into something ZODB can store.

        return value

    def deserialize(self, value):
        # Get raw data and transform it into something more convenient to use.

        return value

    def read(self, parent):
        # Reads some raw data from the parent model and transform it into
        # something more convenient to use. Most of the time, you should use
        # Attribute.deserialize

        return self.deserialize(self.read_raw(parent))

    def write(self, parent, value):
        # Takes user data, transform it into something ZODB can store, and
        # store it into the model parent persistent.

        written_value = self.write_raw(parent, self.serialize(value))
        return self.deserialize(written_value)

    def update(
        self,
        old_value,
        new_value,
        addition=True,
        edition=True,
        deletion=False,
        replacement=False,
    ):
        """Updates the value of the attribute.

        :param old_value: The previous value the attribute had.
        :param new_value: The new value that the attribute should be updated with.
        :param addition: On collections, adds elements to the attribute if they are present in `new_value` and not in `old_value`, defaults to `True`.
        :param edition: On collections, edits the attribute element if present on both `old_value` and `new_value`, defaults to `True`.
        :param deletion: On collections, removes an element if present in `old_value` and absent from `new_value`, default to `False`.
        :param replacement: Replace sub-collections or sub-models instead of editing them in-place, defaults to `False`.
        """
        return new_value if edition else old_value

    def save(self, parent):
        pass

    def delete(self, parent):
        try:
            del parent.mapping[self.key(parent)]
        except KeyError:
            pass

    def index(
        self,
        unique=False,
        key=None,
        values=None,
        search=None,
        mapping=None,
        primary=False,
        nullok=None,
        noneok=None,
    ):
        """
        This method is a shortcut that inits a
        :class:`~sheraf.attributes.index.Index` object on the current attribute.
        It takes the same arguments as :class:`~sheraf.attributes.index.Index` except that by default:

        - the index name will be this attribute name;
        - the mapping parameter will be this attribute `default_index_mapping` parameter;
        - the values parameter will be this attribute :meth:`~sheraf.attributes.Attribute.values` method;
        - the search parameter will be this attribute :meth:`~sheraf.attributes.Attribute.search` method;
        - the noneok parameter will be this attribute `noneok` parameter
        - the nullok parameter will be this attribute `nullok` parameter
        """
        self.indexes[key] = Index(
            self,
            unique=unique,
            key=key,
            values=values or self.values,
            search=search or values or self.search or self.values,
            mapping=mapping or self.default_index_mapping,
            primary=primary,
            nullok=nullok if nullok is not None else self.nullok,
            noneok=noneok if noneok is not None else self.noneok,
        )
        self.lazy = False

        return self

    def values(self, value):
        """
        The default transformation that will be applied when storing data if this attribute is indexed
        but the :func:`~sheraf.attributes.Attribute.index` `values_func` parameter is not provided.

        By default no transformation is applied, and the `value` parameter is returned in a :class:`set`.

        This method can be overload so a custom transformation is applied.
        """
        return {self.serialize(value)}

    def search(self, value):
        """
        The default transformation that will be applied when searching for data if this attribute is indexed
        but the :func:`~sheraf.attributes.Attribute.index` `search_func` parameter is not provided.

        By default this calls :meth:`~sheraf.attributes.Attribute.values`.

        This method can be overload so a custom transformation is applied.
        """
        return self.values(value)

    def on_creation(self, *args, **kwargs):
        """
        Decorator for callbacks to call on an attribute creation. The callback
        will be executed before the attribute is created. If the callback yields,
        the part after the yield will be executed after the creation.

        The callback will be passed the new value of the attribute.

        The callback can be freely named.

        >>> class Cowboy(sheraf.Model):
        ...     table = "old_cowboys_creation"
        ...     age = sheraf.IntegerAttribute()
        ...
        ...     @age.on_creation
        ...     def create_age(self, new):
        ...         print("New cowboy aged of", new)
        ...
        >>> with sheraf.connection():
        ...     george = Cowboy.create(age=50)
        New cowboy aged of 50
        """

        def wrapper(func):
            self.cb_creation.append(func)
            return func

        return wrapper if not args else wrapper(args[0])

    def on_edition(self, *args, **kwargs):
        """
        Decorator for callbacks to call on an attribute edition. The callback
        will be executed before the attribute is edited. If the callback yields,
        the part after the yield will be executed after the update.

        The callback will be passed the old and the new value of the attribute.

        The callback can be freely named.

        >>> class Cowboy(sheraf.Model):
        ...     table = "old_cowboys_edition"
        ...     age = sheraf.IntegerAttribute()
        ...
        ...     @age.on_edition
        ...     def update_age(self, new, old):
        ...         print("I was", old, "years old")
        ...         yield
        ...         print("Now I am", new, "years old")
        ...
        >>> with sheraf.connection():
        ...     george = Cowboy.create(age=50)
        ...     george.age = 51
        I was 50 years old
        Now I am 51 years old
        """

        def wrapper(func):
            self.cb_edition.append(func)
            return func

        return wrapper if not args else wrapper(args[0])

    def on_deletion(self, *args, **kwargs):
        """
        Decorator for callbacks to call on an attribute deletion. The callback
        will be executed before the attribute is deleted. If the callback yields,
        the part after the yield will be executed after the deletion.

        The callback will be passed the old value of the attribute.

        The callback can be freely named.

        >>> class Cowboy(sheraf.Model):
        ...     table = "old_cowboys_deletion"
        ...     age = sheraf.IntegerAttribute()
        ...
        ...     @age.on_deletion
        ...     def delete_age(self, old):
        ...         print("Deleting age of", 50)
        ...
        >>> with sheraf.connection():
        ...     george = Cowboy.create(age=50)
        ...     del george.age
        Deleting age of 50
        """

        def wrapper(func):
            self.cb_deletion.append(func)
            return func

        return wrapper if not args else wrapper(args[0])

    def default(self, func):
        """
        Decorator for a callback to call to generate a default value for
        the attribute.

        >>> class Cowboy(sheraf.Model):
        ...     table = "default_cowboys"
        ...     age = sheraf.IntegerAttribute()
        ...
        ...     @age.default
        ...     def default_age(self):
        ...         return 42
        ...
        >>> with sheraf.connection(commit=True):
        ...     george = Cowboy.create()
        ...     george.age
        42
        """
        self._default_value = func



import crypt
import sheraf


class PasswordAccessor:
    def __init__(self, compare, crypted):
        self.compare = compare
        self.crypted = crypted

    def __str__(self):
        return self.crypted

    def __eq__(self, value):
        return self.compare(value, self.crypted)


class PasswordAttribute(sheraf.Attribute):
    """
    Stores crypted password in the database. Once a password has been crypted,
    it cannot be read back in plain text, however, comparisions are still possible.
    Under the hood, the python :func:`~crypt.crypt` method is used.

    The arguments passed to this attribute are passed to :func:`crypt.mksalt`.

    >>> class Cowboy(sheraf.Model):
    ...     table = "cautious_cowboy"
    ...     email = sheraf.StringAttribute()
    ...     password = sheraf.PasswordAttribute()
    ...
    >>> with sheraf.connection(commit=True):  # doctest: +SKIP
    ...     cowboy = Cowboy.create(email="george@abitbol.com", password="$up3r$3cur3")
    ...     assert cowboy.password == "$up3r$3cur3"
    ...     str(cowboy.password)
    '$6$vQzraMQc3G9Mf/zy$IfPSCPO81IzvGwo6AYoS6K6fen552B22DDz.ARwoxmvyFJ9He7.wVLIWbw0RWEIw/oGUblY9YbGvhwUQbtYEV.'

    """

    def __init__(self, **kwargs):
        self.crypt_args = kwargs
        super().__init__()

    @staticmethod
    def crypt(clear, *args, **kwargs):
        salt = crypt.mksalt(**kwargs)
        return crypt.crypt(clear, salt)

    @staticmethod
    def compare(clear, crypted):
        return crypt.crypt(clear, crypted) == crypted

    def serialize(self, value):
        if value is None:
            return None

        return self.crypt(value, **self.crypt_args)

    def deserialize(self, value):
        if value is None:
            return None

        return PasswordAccessor(self.compare, value)



import itertools
import operator
from collections import OrderedDict
from collections.abc import Iterable, Iterator

from BTrees.OOBTree import OOTreeSet, union, intersection, difference

import sheraf.constants
from sheraf.exceptions import InvalidFilterException, InvalidOrderException
from sheraf.tools.more_itertools import unique_everseen


class QuerySet(object):
    """
    A :class:`~sheraf.queryset.QuerySet` is a collection containing
    :class:`~sheraf.models.Model` instances. Like in a regular :class:`set`,
    objects are unique, but the main difference is that
    :class:`~sheraf.queryset.QuerySet` keeps the insertion order.

    :param iterable: A collection of models. If `iterable` is None, then
        `model` must be set.
    :type iterable: Iterable
    :param model_class: A model class to iterate over. If `model` is None, then
        `iterable` must be set. If both are set, `model_class` is ignored.
    :type reversed_iterable: A Model class
    :param predicate: a callable takes an instance as parameter and return a
        Boolean (if True, the instance will be returned in the iteration). If
        None, everything is returned.
    :type predicate: Predicate
    :param kwargs: A dictionnary containing the values expected from the model
        parameters. If `kwargs` is `{"foo": "bar"}` then the queryset will only
        contains models which attribute `foo` is `"bar"`.

    For the following examples, let us work with a simple **Cowboy** model. For
    the sake of simplicity we use a
    :class:`~sheraf.models.IntOrderedNamedAttributesModel` so the first instance
    created will have id 0, the second will have id 1 and so on...

    >>> class Cowboy(sheraf.IntOrderedNamedAttributesModel):
    ...     table = "queryset_people"
    ...     name = sheraf.SimpleAttribute()
    ...     age = sheraf.SimpleAttribute()
    ...
    >>> with sheraf.connection(commit=True):
    ...     peter = Cowboy.create(name="Peter", age=30)
    ...     steven = Cowboy.create(name="Steven", age=30)
    ...     george = Cowboy.create(name="George Abitbol", age=50)

    :class:`~sheraf.queryset.QuerySet` are mostly created by doing requests on a
    Model with :func:`~sheraf.models.indexation.IndexedModel.all`,
    :func:`~sheraf.models.indexation.IndexedModel.filter` or
    :func:`~sheraf.models.indexation.IndexedModel.order`, but can also be
    initialized with custom data.

    >>> with sheraf.connection():
    ...    a = Cowboy.all() # returns a QuerySet with all Cowboy
    ...    b = QuerySet([peter, steven, george]) # This is an equivalent custom QuerySet
    ...    assert list(a) == list(b)

    :class:`~sheraf.queryset.QuerySet` behave like iterators, and can only be
    consumed once.

    >>> with sheraf.connection():
    ...    everybody = Cowboy.all()
    ...    assert [peter, steven, george] == list(everybody)
    ...    assert [] == everybody

    .. note :: :class:`~sheraf.queryset.QuerySet` can be compared against
        anything iterable, but the comparison will consume the
        :class:`~sheraf.queryset.QuerySet`.

    :class:`~sheraf.queryset.QuerySet` keeps the order of insertion.

    >>> assert QuerySet([peter, steven]) != QuerySet([steven, peter])

    :class:`~sheraf.queryset.QuerySet` supports slicing. Slices returns another
    :class:`~sheraf.queryset.QuerySet`.

    >>> with sheraf.connection():
    ...     assert peter == Cowboy.all()[0]
    ...     assert QuerySet([peter, steven]) == Cowboy.all()[0:2]
    ...     assert QuerySet([peter, steven, george]) == Cowboy.all()[0:]
    """

    def __init__(self, iterable=None, model_class=None, predicate=None, **kwargs):
        self.filters = OrderedDict(kwargs)
        self._iterable = iterable
        self._iterator = None
        self._predicate = predicate
        self._start = None
        self._stop = None
        self._step = None
        self.model = model_class
        self.orders = OrderedDict()

        if iterable is None and model_class is None:
            self._iterable = []

        self.__iter__()

    def __iter__(self):
        return self

    def __repr__(self):
        result = "<QuerySet"
        if self.model:
            result = f"{result} model={self.model.__name__}"

        if self._iterable:
            result = f"{result} iterable={self._iterable}"

        if self._predicate:
            result = f"{result} predicate={self._predicate}"

        if self.filters:
            filters = ", ".join(f"{k}={v[1]}" for k, v in self.filters.items())
            result = f"{result} filters=({filters})"

        result = f"{result}>"
        return result

    def __next__(self):
        if not self._iterator:
            self._init_iterator()

        return next(self._iterator)

    def __eq__(self, other):
        if isinstance(other, Iterable):
            return all(
                a == b
                for a, b in itertools.zip_longest(
                    iter(self), iter(other), fillvalue=object()
                )
            )
        return super().__eq__(other)

    def __and__(self, other):
        return QuerySet(intersection(OOTreeSet(self), OOTreeSet(other)))

    def __or__(self, other):
        return QuerySet(union(OOTreeSet(self), OOTreeSet(other)))

    def __xor__(self, other):
        return QuerySet(difference(OOTreeSet(self), OOTreeSet(other)))

    def __add__(self, other):
        return QuerySet(unique_everseen(itertools.chain(self, other)))

    def __bool__(self):
        try:
            next(self.copy())
            self._iterator = None
            return True
        except StopIteration:
            return False

    def __getitem__(self, item):
        qs = self.copy()

        if isinstance(item, slice):
            qs._start, qs._stop, qs._step = item.start, item.stop, item.step
            return qs

        else:
            qs._start, qs._stop, qs._step = item, item + 1, 1
            return next(qs)

    def __len__(self):
        """
        :return: The number of objects in the
            :class:`~sheraf.queryset.QuerySet`.

        >>> with sheraf.connection():
        ...     assert Cowboy.create()
        ...     qs = Cowboy.all()
        ...     assert len(qs) == 1
        """
        # No shortcut possible when there is a predicate,
        # or when there are several filters
        if self._predicate or not self.model or len(self.filters) > 1:
            return sum(1 for _ in self.copy())

        # We basically want to search all the models
        if not self.filters:
            return self.model.count()

        # Take the unique filter, and hope it is indexed
        index_name = list(self.filters.keys())[0]
        _, filter_value, filter_search_func = list(self.filters.values())[0]
        if index_name not in self.model.indexes:
            return sum(1 for _ in self.copy())

        index = self.model.indexes[index_name]
        if filter_search_func:
            index_values = index.details.call_search_func(self.model, filter_value)
        else:
            index_values = filter_value

        if index.details.unique:
            return sum(int(index.has_item(v)) for v in index_values)

        else:
            return sum(
                len(index.get_item(v)) for v in index_values if index.has_item(v)
            )

    @property
    def indexed_filters(self):
        return [
            (name, value, search_func, self.orders.get(name) == sheraf.constants.DESC,)
            for (name, value, search_func) in self.filters.values()
            if name in self.model.indexes
        ]

    @property
    def non_indexed_filters(self):
        return [
            (name, value)
            for (name, value, _) in self.filters.values()
            if name not in self.model.indexes and name in self.model.attributes
        ]

    def _index_keys(self, index, value, search_func):
        return (
            index.details.call_search_func(self.model, value)
            if search_func
            else [value]
        )

    def _objects_ids(self, name, value, search_func, reverse):
        index = self.model.indexes[name]
        keys = self._index_keys(index, value, search_func)

        if index.details.unique:
            mappings = (index.get_item(key, True) for key in keys)

        else:
            mappings_lists = (index.get_item(key, True) for key in keys)
            mappings = (m for l in mappings_lists if l for m in l)

        return (m[self.model.primary_key()] for m in mappings if m)

    def _multi_indexes_iterator(self):
        ids_sets = (set(self._objects_ids(*index)) for index in self.indexed_filters)
        raw_ids = set.intersection(*ids_sets)

        pk_attribute = self.model.attributes[self.model.primary_key()]
        ids = (pk_attribute.deserialize(id_) for id_ in raw_ids)
        return ids

    def _mono_indexes_iterator(self):
        pk_attribute = self.model.attributes[self.model.primary_key()]
        raw_ids = self._objects_ids(*self.indexed_filters[0])
        unique_raw_ids = unique_everseen(raw_ids)
        ids = (pk_attribute.deserialize(id_) for id_ in unique_raw_ids)
        return ids

    def _primary_index_iterator(self):
        identifier_index = self.model.indexes[self.model.primary_key()]
        reverse = self.orders.get(self.model.primary_key()) == sheraf.constants.DESC
        ids = identifier_index.iterkeys(reverse)
        return ids

    def _init_iterator(self):
        if self._iterable:
            iterator = iter(self._iterable)

        elif not self.model:
            iterator = iter([])

        # iterate all items on the primary key
        elif not self.indexed_filters:
            iterator = self._primary_index_iterator()

        # iterator on several indexed filters
        elif not self.orders:
            iterator = self._mono_indexes_iterator()

        if self.model:
            iterator = self.model.read_these(iterator)

        # Checks the models fits all the filters
        iterator = (
            model for model in iterator if self._model_has_expected_values(model)
        )

        # Successively sorts the list from the less important
        # order to the most important order.
        if self.orders:
            iterable = iterator
            for attribute, order in reversed(self.orders.items()):
                iterable = sorted(
                    iterable,
                    key=operator.attrgetter(attribute),
                    reverse=(order == sheraf.constants.DESC),
                )
            iterator = iter(iterable)

        # Only select a slice of the wanted models
        if self._start is not None or self._stop is not None or self._step is not None:
            iterator = itertools.islice(iterator, self._start, self._stop, self._step)

        self._iterator = iterator

    def _model_has_expected_values(self, model):
        if not all(
            getattr(model, filter_name) == expected_value
            for filter_name, expected_value in self.non_indexed_filters
        ):
            return False

        if not all(
            (
                set(model.indexes[name].details.call_search_func(model, value))
                & set(model.indexes[name].details.get_model_values(model))
            )
            if search_func
            else (value in model.indexes[name].details.get_model_values(model))
            for name, value, search_func, _ in self.indexed_filters
        ):
            return False

        return not self._predicate or self._predicate(model)

    def count(self):
        return len(self)

    def copy(self):
        """Copies the :class:`~sheraf.queryset.QuerySet` without consuming it.

        >>> with sheraf.connection():
        ...     peter = Cowboy.create(name="Peter")
        ...     steven = Cowboy.create(name="Steven")
        ...     george = Cowboy.create(name="George")
        ...     qall = Cowboy.all()
        ...     qcopy = qall.copy()
        ...
        ...     assert [peter, steven, george] == qall
        ...     # now qall is consumed
        ...
        ...     assert [peter, steven, george] == qcopy
        ...     # now qcopy is consumed
        """

        if isinstance(self._iterable, Iterator):
            self._iterable, iterable = itertools.tee(self._iterable)
        else:
            iterable = self._iterable

        qs = QuerySet(iterable, self.model)
        qs.filters = self.filters.copy()
        qs.orders = self.orders.copy()
        qs._predicate = self._predicate
        return qs

    def delete(self):
        """Delete the objects contained in the queryset.

        Avoids problems when itering on deleted objects.
        """
        identifiers = [(m.__class__, m.identifier) for m in self]
        for klass, identifier in identifiers:
            klass.read(identifier).delete()

    def filter(self, predicate=None, **kwargs):
        """Refine a copy of the current :class:`~sheraf.queryset.QuerySet` with
        further tests.

        :param predicate: filter instance by returning a truthy value.
            If `None` everything is selected.
        :type predicate: callable object
        :param kwargs: A dictionnary containing the values expected from the
            model parameters. If ``kwargs`` is ``{"foo": "bar"}`` then the
            queryset will only contains models which attribute ``foo``
            is ``"bar"``.
        :type kwargs: A dictionary which keys must be valid attributes of the
            model iterated.
        :return: A copy of the current :class:`~sheraf.queryset.QuerySet`
            refined with further tests.
        :return type: :class:`~sheraf.queryset.QuerySet`

        It is possible to chain :func:`~sheraf.queryset.QuerySet.filter` calls:

        >>> with sheraf.connection():
        ...    assert Cowboy.filter(name="George Abitbol", age=50) == \\
        ...           Cowboy.filter(name="George Abitbol").filter(age=50)
        ...
        >>> with sheraf.connection():
        ...    assert Cowboy.filter(lambda person: "Abitbol" in person.name, age=50) == \\
        ...           Cowboy.filter(lambda person: "Abitbol" in person.name).filter(age=50)

        An attribute cannot be filtered twice:

        >>> with sheraf.connection():
        ...    Cowboy.filter(age=30).filter(age=40)
        Traceback (most recent call last):
            ...
        sheraf.exceptions.InvalidFilterException: Some filter parameters appeared twice

        .. note::   Filtering on indexed attributes is more performant than filtering on non-indexed attributes. See :func:`~sheraf.attributes.Attribute.index`.
        """
        return self._filter(False, predicate=predicate, **kwargs)

    def search(self, **kwargs):
        """
        Refine a copy of the current :class:`~sheraf.queryset.QuerySet` with further tests.

        This method is very similar to :func:`~sheraf.queryset.QuerySet.filter` except the
        values it takes are transformed with the same way values are transformed at indexation.
        TODO: pas très clair

        For instance, if an attribute indexes its values with a lowercase search_func, the
        :func:`~sheraf.queryset.QuerySet.search` attributes will go through the same
        search_func. Hence it allows to pass uppercase filter values, while
        :func:`~sheraf.queryset.QuerySet.filter` does not allow this.

        >>> class MyCustomModel(sheraf.Model):
        ...     table = "my_custom_model"
        ...     my_attribute = sheraf.SimpleAttribute().index(
        ...        values=lambda string: {string.lower()}
        ...     )
        ...
        >>> with sheraf.connection(commit=True):
        ...     m = MyCustomModel.create(my_attribute="FOO")
        ...
        >>> with sheraf.connection():
        ...     assert [m] == MyCustomModel.search(my_attribute="foo")
        ...     assert [m] == MyCustomModel.filter(my_attribute="foo")
        ...
        ...     assert [m] == MyCustomModel.search(my_attribute="FOO")
        ...     assert [] == MyCustomModel.filter(my_attribute="FOO")
        """

        return self._filter(True, **kwargs)

    def _filter(self, search_func, predicate=None, **kwargs):
        qs = self.copy()
        if self.model:
            for filter_name in kwargs.keys():
                if (
                    filter_name not in self.model.attributes
                    and filter_name not in self.model.indexes
                ):
                    raise sheraf.exceptions.InvalidFilterException(
                        "{} has no attribute {}".format(
                            self.model.__name__, filter_name
                        )
                    )
        kwargs_values = OrderedDict(
            {
                filter_name: (filter_name, filter_value, search_func)
                for filter_name, filter_value in kwargs.items()
            }
        )
        common_attributes = set(qs.filters) & set(kwargs_values)
        invalid_common_attributes = any(
            key for key in common_attributes if qs.filters[key] != kwargs_values[key]
        )
        if invalid_common_attributes:
            raise InvalidFilterException("Some filter parameters appeared twice")

        qs.filters.update(kwargs_values)

        if not qs._predicate:
            qs._predicate = predicate

        elif predicate:
            old_predicate = qs._predicate
            qs._predicate = lambda m: old_predicate(m) and predicate(m)

        return qs

    def order(self, *args, **kwargs):
        """Copies the current :class:`~sheraf.queryset.QuerySet` and adds more
        order to it.

        :param args: There can be only one positionnal argument. Choose to iterate
            over ids in an ascending or a descending way.
        :type args: ``sheraf.ASC`` or ``sheraf.DESC``
        :param kwargs: Further parameters will set an order on the matching
            model attributes.
        :type kwargs: A dictionary which keys must be valid attributes of the
            model iterated, and the values must be ``sheraf.ASC`` or
            ``sheraf.DESC``
        :return: A copy of the current :class:`~sheraf.queryset.QuerySet` with
            refined order.
        :return type: :class:`~sheraf.queryset.QuerySet`

        The default order is the ascending model ids.

        >>> with sheraf.connection(commit=True):
        ...     peter = Cowboy.create(name="Peter", age=35)
        ...     steven = Cowboy.create(name="Steven", age=35)
        ...     george = Cowboy.create(name="George", age=50)
        ...
        >>> with sheraf.connection():
        ...     assert [peter, steven, george] == Cowboy.all()
        ...     assert [peter, steven, george] == Cowboy.all().order(sheraf.ASC)
        ...     assert [george, steven, peter] == Cowboy.all().order(sheraf.DESC)
        ...
        ...     assert [george, peter, steven] == Cowboy.all().order(name=sheraf.ASC)
        ...     assert [steven, peter, george] == Cowboy.all().order(name=sheraf.DESC)

        Several order parameters can be passed, either as arguments of the
        function, or by calling :func:`~sheraf.queryset.QuerySet.order` calls.

        >>> with sheraf.connection():
        ...     assert [george, peter, steven] == Cowboy.all().order(age=sheraf.DESC, name=sheraf.ASC)

        .. note:: Sorting on indexed attributes is more performant than
            sorting on other attributes. See
            :func:`~sheraf.attributes.Attribute.index`.
            The less :func:`~sheraf.queryset.QuerySet.order` parameters are
            passed, the better performances will be.
        """
        if not self.model and (args or not kwargs):
            raise InvalidOrderException(
                "QuerySets without models should have explict order arguments"
            )

        if len(args) > 1:
            raise InvalidOrderException(
                "Only one 'order' positionnal parameter is allowed."
            )

        qs = self.copy()

        if self.model:
            for attribute, value in kwargs.items():
                if attribute not in self.model.attributes:
                    raise sheraf.exceptions.InvalidOrderException(
                        "{} has no attribute {}".format(self.model.__name__, attribute)
                    )

                if value not in (sheraf.constants.ASC, sheraf.constants.DESC):
                    raise sheraf.exceptions.InvalidOrderException(
                        "Parameter {} has an invalid order value {}".format(
                            attribute, value
                        )
                    )

        if len(args) > 0:
            identifier = args[0]
            if identifier not in (sheraf.constants.ASC, sheraf.constants.DESC):
                raise InvalidOrderException(
                    "Parameter id has an invalid order value {}".format(identifier)
                )

            if self.model.primary_key() in qs.orders:
                raise InvalidOrderException("Id order has been set twice")

            qs.orders[self.model.primary_key()] = identifier

        common_attributes = set(qs.orders) & set(kwargs)
        if common_attributes:
            raise InvalidOrderException("Some order parameters appeared twice")

        qs.orders.update(kwargs)

        return qs

    def get(self):
        """If the :class:`~sheraf.queryset.QuerySet` contains one, and only one
        item, this method returns the item. If the
        :class:`~sheraf.queryset.QuerySet` contains several objects, it raises a
        :class:`~sheraf.exceptions.QuerySetUnpackException`. If the
        :class:`~sheraf.queryset.QuerySet` is empty, it raises a
        :class:`~sheraf.exceptions.EmptyQuerySetUnpackException`.

        >>> with sheraf.connection():
        ...     peter = Cowboy.create(name="Peter")
        ...     steven = Cowboy.create(name="Steven")
        ...     assert peter == Cowboy.filter(name="Peter").get()
        ...     Cowboy.all().get()
        Traceback (most recent call last):
            ...
        sheraf.exceptions.QuerySetUnpackException: Trying to unpack more than 1 value from a QuerySet
        >>> with sheraf.connection():
        ...     Cowboy.filter(age=30).get()
        Traceback (most recent call last):
            ...
        sheraf.exceptions.EmptyQuerySetUnpackException: Trying to unpack an empty QuerySet
        >>> with sheraf.connection():
        ...     Cowboy.filter(name="Unknown cowboy").get()
        Traceback (most recent call last):
            ...
        sheraf.exceptions.EmptyQuerySetUnpackException: Trying to unpack an empty QuerySet
        """
        try:
            element = next(self)
        except StopIteration:
            raise sheraf.exceptions.EmptyQuerySetUnpackException()

        try:
            next(self)
        except StopIteration:
            return element
        else:
            raise sheraf.exceptions.QuerySetUnpackException()



import sheraf


class EnumAccessor:
    def __init__(self, enum):
        self.enum = enum

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if (
                name.startswith("is_")
                and name.upper()[3:] in self.enum.__class__.__members__
            ):
                return self.value == self.enum.__class__.__members__[name.upper()[3:]]
            return getattr(self.enum, name)

    def __eq__(self, other):
        return self.enum.value == other

    def __lt__(self, other):
        if isinstance(other, EnumAccessor):
            return self.enum.value < other.value
        return self.enum.value < other

    def __le__(self, other):
        if isinstance(other, EnumAccessor):
            return self.enum.value <= other.value
        return self.enum.value <= other

    def __gt__(self, other):
        if isinstance(other, EnumAccessor):
            return self.enum.value > other.value
        return self.enum.value > other

    def __ge__(self, other):
        if isinstance(other, EnumAccessor):
            return self.enum.value >= other.value
        return self.enum.value >= other

    def __hash__(self):
        return hash(self.enum)

    def __str__(self):
        return str(self.enum.value)


class EnumAttribute(sheraf.Attribute):
    """
    Takes an :class:`~enum.Enum` and an optional :class:`~sheraf.attributes.Attribute`, filters the data with the :class:`~enum.Enum` and serialize it with the methods of the :class:`~sheraf.attributes.Attribute`.

    >>> import enum
    ...
    >>> class Cowboy(sheraf.Model):
    ...     table = "enum_cowboys"
    ...
    ...     class Status(enum.IntEnum):
    ...         FARMER = 0
    ...         COWBOY = 1
    ...         SHERIF = 2
    ...
    ...     status = sheraf.EnumAttribute(Status, sheraf.IntegerAttribute())
    ...
    >>> with sheraf.connection(commit=True):
    ...     george = Cowboy.create(status=Cowboy.Status.SHERIF)
    ...
    ...     assert george.status == 2
    ...     assert george.status == Cowboy.Status.SHERIF

    .. note :: For every value ``FOOBAR`` defined in the enum, the attributes
               defines an accessor ``is_foobar``.

    >>> with sheraf.connection(commit=True):
    ...     assert george.status.is_sherif
    ...     assert not george.status.is_farmer
    """

    def __init__(self, enum, attribute=None, **kwargs):
        self.attribute = attribute or sheraf.Attribute
        self.enum = enum
        super().__init__(**kwargs)

    def serialize(self, value):
        if value is None:
            return None

        if not isinstance(value, EnumAccessor):
            value = self.attribute.serialize(value)

        enum_value = self.enum(value).value
        return self.attribute.serialize(enum_value)

    def deserialize(self, value):
        if value is None:
            return None

        data = self.attribute.deserialize(value)
        return EnumAccessor(self.enum(data))



from BTrees.IOBTree import IOBTree


class LargeList(IOBTree):
    """Large List."""

    LENGTH_KEY = -1

    def __init__(self, items=None):
        items = items if items is not None else []
        self.extend(items)

    def append(self, item, unique=False):
        if unique and item in self:
            return

        new_index = len(self)
        self._set_length(len(self) + 1)
        self[new_index] = item

    def __eq__(self, other):
        if len(self) != len(other):
            return False

        return all(mine == their for mine, their in zip(self, other))

    def __len__(self):
        return IOBTree.get(self, self.LENGTH_KEY, self.LENGTH_KEY + 1)

    def __add__(self, other):
        return LargeList(list(self) + other)

    def _set_length(self, length):
        IOBTree.__setitem__(self, self.LENGTH_KEY, length)

    def extend(self, items, unique=False):
        for _item in items:
            self.append(_item, unique)

    def insert(self, indice, element):
        self._set_length(len(self) + 1)
        for i in range(len(self) - 1, indice, -1):
            self[i] = self[i - 1]
        self[indice] = element

    def remove(self, item, all=False):
        found = False
        for key, value in self.items():
            if key == self.LENGTH_KEY:
                continue

            if item != value:
                continue

            found = True
            IOBTree.__delitem__(self, key)
            for i in range(key + 1, len(self)):
                item = IOBTree.__getitem__(self, i)
                IOBTree.__setitem__(self, i - 1, item)
                IOBTree.__delitem__(self, i)
            self._set_length(len(self) - 1)

            if not all:
                return

        if not all or not found:
            raise ValueError("{} not in {}".format(item, self))

    def pop(self):
        length = len(self)
        self._set_length(length - 1)
        return super().pop(length - 1)

    def __iter__(self):
        _values = IOBTree.itervalues(self)
        try:
            next(_values)
            return _values
        except StopIteration:
            return iter({})

    def __contains__(self, item):
        return item in iter(self)

    def __getitem__(self, item):
        if isinstance(item, slice):
            _slice = self._reslice(item)
            return (
                IOBTree.__getitem__(self, index)
                for index in range(
                    _slice.start, min(len(self), _slice.stop), _slice.step
                )
            )

        elif isinstance(item, int):
            if item >= len(self):
                raise IndexError

            try:
                if not item:
                    return IOBTree.__getitem__(self, 0)
                return IOBTree.__getitem__(self, item > 0 and item or len(self) + item)
            except KeyError as ex:
                raise IndexError(ex)

        else:
            raise TypeError(
                f"Invalid LargeList key '{item}'. It must be an integer or a slice."
            )

    def __setitem__(self, key, value):
        if not isinstance(key, slice) and key >= len(self):
            raise IndexError
        return super().__setitem__(key, value)

    def _reslice(self, item):
        if item.step is None or item.step > 0:
            return slice(self._start(item), self._stop(item), self._step(item))
        else:
            return slice(self._start_down(item), self._stop_down(item), item.step)

    def _start(self, item):
        if item.start is None:
            return 0
        elif item.start < 0:
            return len(self) + item.start
        else:
            return item.start

    def _stop(self, item):
        if item.stop is None:
            return len(self)
        elif item.stop < 0:
            return len(self) + item.stop
        else:
            return item.stop

    def _step(self, item):
        return item.step or 1

    def _start_down(self, item):
        if item.start is None:
            return len(self) - 1
        elif item.start < 0:
            return len(self) + item.start
        else:
            return item.start

    def _stop_down(self, item):
        if item.stop is None:
            return -1
        elif item.stop < 0:
            return len(self) + item.stop
        else:
            return item.stop



from sheraf.attributes import Attribute

from .attributes import NamedAttributesModel


class InlineModel(NamedAttributesModel):
    """
    :class:`~sheraf.models.inline.InlineModel` behaves like a regular model, but
    it is not indexed by itself. This has several consequences:
    - The ``table`` attribute is not needed.
    - The :func:`~sheraf.models.base.BaseModel.read` method is not available.
    :class:`~sheraf.models.inline.InlineModel` aims to be used in combination
    with :class:`~sheraf.attributes.models.InlineModelAttribute`.

    >>> class Gun(sheraf.InlineModel):
    ...     nb_bullets = sheraf.IntegerAttribute()
    ...
    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.SimpleAttribute()
    ...     gun = sheraf.InlineModelAttribute(Gun)
    ...
    >>> with sheraf.connection(commit=True):
    ...     george = Cowboy.create(
    ...         name="George Abitbol",
    ...         gun=Gun.create(nb_bullets=5),
    ...     )
    ...     assert 5 == george.gun.nb_bullets

    You can manage your own indexation by combining
    :class:`~sheraf.attributes.models.InlineModelAttribute` with a collection
    attribute, like :class:`~sheraf.attributes.collections.DictAttribute`.

    >>> class Gun(sheraf.InlineModel):
    ...     nb_bullets = sheraf.IntegerAttribute()
    ...
    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.SimpleAttribute()
    ...     guns = sheraf.LargeDictAttribute(
    ...         sheraf.InlineModelAttribute(Gun)
    ...     )
    ...
    >>> with sheraf.connection(commit=True):
    ...     george = Cowboy.create(
    ...         name="George Abitbol",
    ...         guns = {
    ...             "martha": Gun.create(nb_bullets=6),
    ...             "gretta": Gun.create(nb_bullets=5),
    ...         }
    ...     )
    ...     assert 6 == george.guns["martha"].nb_bullets

    :class:`~sheraf.models.inline.InlineModel` can also be anonymous. To create
    an anonymous model instance, just pass the attributes list as parameter of the
    class.

    >>> class Cowboy(sheraf.Model):
    ...     table = "cowboy"
    ...     name = sheraf.SimpleAttribute()
    ...     horse = sheraf.InlineModelAttribute(
    ...         sheraf.InlineModel(
    ...             name=sheraf.SimpleAttribute()
    ...         )
    ...     )
    ...
    >>> with sheraf.connection(commit=True):
    ...     george = Cowboy.create(horse=dict(name="Jolly Jumper"))
    ...     george.horse.name
    'Jolly Jumper'
    """

    def __init__(self, **kwargs):
        if kwargs:
            self.__class__.attributes = {}

        super().__init__()

        for k, v in kwargs.items():
            if isinstance(v, Attribute):
                v.attribute_name = k
                self.attributes[k] = v



import itertools
import types
import warnings

import sheraf.exceptions
from sheraf.models.base import BaseModel, BaseModelMetaclass
from sheraf.models.indexmanager import (
    SimpleIndexManager,
    MultipleDatabaseIndexManager,
)


class BaseIndexedModelMetaclass(BaseModelMetaclass):
    def __new__(cls, name, bases, attrs):
        klass = super().__new__(cls, name, bases, attrs)
        klass.indexes = {}

        def add_index(name, index, attributes, add_to_attribute=True):
            # Get the real attributes objects when string have been passed as attributes
            # in the index.

            index.key = index.key or name
            if not index.attributes:
                raise sheraf.exceptions.SherafException(
                    f"The {index.key} index must have at least one attribute."
                )

            new_attrs = []
            for a in index.attributes:
                if isinstance(a, str):
                    try:
                        a = attributes[a]
                    except KeyError:
                        raise sheraf.exceptions.SherafException(
                            f"The {index.key} index has a wrong attribute with name '{a}'."
                        )
                if not isinstance(a, sheraf.Attribute):
                    raise sheraf.exceptions.SherafException(
                        f"The {index.key} index has a wrong attribute."
                    )

                new_attrs.append(a)
            index.attributes = new_attrs

            # Get the attributes from the attribute names
            index.values_funcs = {
                func: [
                    [
                        attributes[attr] if isinstance(attr, str) else attr
                        for attr in attrs
                    ]
                    for attrs in attrs_groups
                ]
                for func, attrs_groups in index.values_funcs.items()
            }

            for attribute in index.attributes:
                if add_to_attribute:
                    attribute.indexes[index.key] = index

                attribute.lazy = False

            # Assign the default values func to attributes without
            # values func.
            attrs_with_func = [
                attr
                for func, attr_groups in index.values_funcs.items()
                if func is not index.default_values_func
                for attrs in attr_groups
                for attr in attrs
            ]

            index.values_funcs[index.default_values_func] = [
                [attribute]
                for attribute in index.attributes
                if attribute not in attrs_with_func
            ]

            klass.indexes[index.key] = klass.index_manager(index)

        for name, index in attrs.items():
            if isinstance(index, sheraf.attributes.index.Index):
                add_index(name, index, klass.attributes)

        for attribute in klass.attributes.values():
            for index_key, index in attribute.indexes.items():
                add_index(attribute.key(klass), index, klass.attributes, False)

        return klass


class BaseIndexedModel(BaseModel, metaclass=BaseIndexedModelMetaclass):
    """
    This class handles the whole indexation mechanism. The mechanisms
    for reading or iterating over models in the database are handled
    here.
    """

    _primary_key = None
    _is_first_instance = None

    def __init__(self, *args, **kwargs):
        self._identifier = None
        self._raw_identifier = None

        if not self.primary_key():
            raise sheraf.exceptions.PrimaryKeyException(
                "{} inherit from IndexedModel but has no primary key. Cannot create.".format(
                    self.__class__.__name__
                )
            )

        super().__init__(*args, **kwargs)

    @classmethod
    def primary_key(cls):
        if cls._primary_key is None:
            for index_name, index in cls.indexes.items():
                if not index.details.primary:
                    continue

                if cls._primary_key is None:
                    cls._primary_key = index_name

                else:
                    raise sheraf.exceptions.PrimaryKeyException(
                        "A model can have only one primary key. '{}' has '{}' and '{}'".format(
                            cls.__class__.__name__,
                            cls._primary_key,
                            index_name,
                        )
                    )

        return cls._primary_key

    @classmethod
    def all(cls):
        """
        :return: A :class:`~sheraf.queryset.QuerySet` containing all the
            registered models.
        """
        return sheraf.queryset.QuerySet(model_class=cls)

    def initialize(self, **kwargs):
        if self.primary_key() in kwargs:
            identifier = kwargs[self.primary_key()]
            del kwargs[self.primary_key()]
        else:
            identifier = self.attributes[self.primary_key()].create(self)

        self.__setattr__(self.primary_key(), identifier)

        super().initialize(**kwargs)

    @classmethod
    def _check_args(cls, *args, **kwargs):
        if len(args) + len(kwargs) != 1:
            raise TypeError(
                "BaseIndexedModel.read (and variants) take only one positionnal or named parameter"
            )

        if args:
            index_name = cls.primary_key()
            key = args[0]

        else:
            index_name, key = list(kwargs.items())[0]

        try:
            index = cls.indexes[index_name]
        except KeyError:
            raise sheraf.exceptions.InvalidIndexException(
                "'{}' is not a valid index".format(index_name)
            )

        return index, key

    @classmethod
    def read(cls, *args, **kwargs):
        """
        Get a model instance from its identifier. If the model identifier is not valid, a
        :class:`~sheraf.exceptions.ModelObjectNotFoundException` is raised.

        The function takes only one parameter which key is the index where to
        search, and which value is the index identifier. If the index is
        multiple, a :class:`~sheraf.exceptions.MultipleIndexException` is
        raised.
        By default the index used is the `id` index.

        :param *args*: The ``identifier`` of the model. There can be only one positionnal or
                      keyword argument.
        :param *kwargs*: The ``identifier`` of the model. There can be only one positionnal or
                        keyword argument.
        :return: The :class:`~sheraf.models.indexation.BaseIndexedModel` matching the id.

        >>> class MyModel(sheraf.Model):
        ...     table = "my_model"
        ...     unique = sheraf.SimpleAttribute().index(unique=True)
        ...     multiple = sheraf.SimpleAttribute().index()
        ...
        >>> with sheraf.connection():
        ...     m = MyModel.create(unique="A", multiple="B")
        ...     assert MyModel.read(m.id) == m
        ...     assert MyModel.read(unique="A") == m
        ...
        >>> with sheraf.connection():
        ...     MyModel.read("invalid")
        Traceback (most recent call last):
            ...
        ModelObjectNotFoundException
        >>> with sheraf.connection():
        ...     MyModel.read(multiple="B")
        Traceback (most recent call last):
            ...
        MultipleIndexException
        """

        index, key = cls._check_args(*args, **kwargs)

        if not index.details.unique:
            raise sheraf.exceptions.MultipleIndexException(
                "'{}' is a multiple index and cannot be used with 'read'".format(
                    index.details.key
                )
            )

        return cls._decorate(cls._read_model_index(key, index))

    @classmethod
    def read_these(cls, *args, **kwargs):
        """
        Get model instances from their identifiers. Unlike
        :func:`~sheraf.models.indexation.BaseModel.read_these`,If an instance
        identifiers does not exist, a :class:`~sheraf.exceptions.ModelObjectNotFoundException`
        is raised.

        The function takes only one parameter which key is the index where to
        search, and which values are the index identifier.
        By default the index used is the `id` index.

        :return: A generator over the models matching the keys.

        >>> class MyModel(sheraf.IntIndexedNamedAttributesModel):
        ...     table = "my_model"
        ...
        >>> with sheraf.connection():
        ...     m1 = MyModel.create(id=1)
        ...     m2 = MyModel.create(id=2)
        ...
        ...     assert [m1, m2] == list(MyModel.read_these([m1.id, m2.id]))
        ...     list(MyModel.read_these(["invalid"]))
        Traceback (most recent call last):
            ...
        sheraf.exceptions.ModelObjectNotFoundException: Id 'invalid' not found in MyModel, 'id' index
        """

        index, keys = cls._check_args(*args, **kwargs)

        if index.details.unique:
            return (cls._decorate(cls._read_model_index(key, index)) for key in keys)

        else:
            return itertools.chain.from_iterable(
                (
                    cls._decorate(mapping)
                    for mapping in cls._read_model_index(key, index)
                )
                for key in keys
            )

    @classmethod
    def read_these_valid(cls, *args, **kwargs):
        """
        Return model instances from an index. Unlike :func:`~sheraf.models.indexation.BaseModel.read_these`,
        invalid index values are ignored.

        The function takes only one parameter which key is the index where to
        search, and which values are the index identifier.
        By default the index used is the `id` index.

        :return: A generator over the models matching the keys.

        >>> class MyModel(sheraf.IntIndexedNamedAttributesModel):
        ...     table = "my_model"
        ...
        >>> with sheraf.connection():
        ...     m1 = MyModel.create(id=1)
        ...     m2 = MyModel.create(id=2)
        ...
        ...     assert [m1, m2] == list(MyModel.read_these_valid([m1.id, m2.id]))
        ...     assert [m1, m2] == list(MyModel.read_these_valid([m1.id, 42, m2.id]))
        """

        index, keys = cls._check_args(*args, **kwargs)

        if index.details.unique:
            return (
                cls._decorate(index.get_item(key))
                for key in keys
                if index.has_item(key)
            )

        else:
            return itertools.chain.from_iterable(
                (cls._decorate(mapping) for mapping in index.get_item(key))
                for key in keys
                if index.has_item(key)
            )

    @classmethod
    def _read_model_index(cls, key, index):
        try:
            return index.get_item(key)
        except KeyError:
            raise sheraf.exceptions.ModelObjectNotFoundException(
                cls, key, index.details.key
            )

    @classmethod
    def index_table_rebuild(cls, index_names=None):
        """
        Resets a model indexation tables.

        This method should be called if an attribute became indexed in an already
        populated database.

        :param index_names: A list of index names to reset. If `None`, all the
                            indexes will be reseted. The primary index cannot be
                            resetted.
        """
        if not index_names:
            indexes = cls.indexes.values()
        else:
            indexes = [
                index
                for index_name, index in cls.indexes.items()
                if index_name in index_names
            ]

        for index in indexes:
            if index.details.primary:
                continue

            index.delete()

        for m in cls.all():
            for index in indexes:
                if not index.details.primary:
                    index.add_item(m)

    @classmethod
    def filter(cls, predicate=None, **kwargs):
        """Shortcut for :func:`sheraf.queryset.QuerySet.filter`.

        :return: :class:`sheraf.queryset.QuerySet`
        """
        return sheraf.queryset.QuerySet(model_class=cls).filter(
            predicate=predicate, **kwargs
        )

    @classmethod
    def search(cls, *args, **kwargs):
        """
        Shortcut for :func:`sheraf.queryset.QuerySet.search`.

        :return: :class:`sheraf.queryset.QuerySet`
        """
        return sheraf.queryset.QuerySet(model_class=cls).search(*args, **kwargs)

    @classmethod
    def order(cls, *args, **kwargs):
        """Shortcut for :func:`sheraf.queryset.QuerySet.order`.

        :return: :class:`sheraf.queryset.QuerySet`
        """
        return sheraf.queryset.QuerySet(model_class=cls).order(*args, **kwargs)

    @classmethod
    def get(cls, *args, **kwargs):
        """Shortcut for :func:`sheraf.queryset.QuerySet.filter` and
        :func:`sheraf.queryset.QuerySet.get`. ``Cowboy.get(name="Peter")`` and
        ``Cowboy.filter(name="Peter").get()`` are equivalent.

        :return: The instance of the model if the filter matches exactly one
            instance. Otherwise, it raises a
            :class:`~sheraf.exceptions.QuerySetUnpackException`.

        >>> class Cowboy(sheraf.Model):
        ...     table = "people"
        ...     name = sheraf.SimpleAttribute()
        ...     age = sheraf.SimpleAttribute()
        ...
        >>> with sheraf.connection(commit=True):
        ...     peter = Cowboy.create(name="Peter", age=30)
        ...     steven = Cowboy.create(name="Steven", age=30)
        ...     assert peter == Cowboy.get(name="Peter")
        ...
        >>> with sheraf.connection():
        ...     Cowboy.get()
        Traceback (most recent call last):
            ...
        sheraf.exceptions.QuerySetUnpackException: Trying to unpack more than 1 value from a QuerySet
        >>> with sheraf.connection():
        ...     Cowboy.get(age=30)
        Traceback (most recent call last):
            ...
        sheraf.exceptions.QuerySetUnpackException: Trying to unpack more than 1 value from a QuerySet
        >>> with sheraf.connection():
        ...     Cowboy.get(name="Unknown cowboy")
        Traceback (most recent call last):
            ...
        sheraf.exceptions.EmptyQuerySetUnpackException: Trying to unpack an empty QuerySet
        """
        return cls.filter(*args, **kwargs).get()

    def __repr__(self):
        try:
            identifier = (
                self.identifier
                if self.mapping is not None and self.primary_key() in self.mapping
                else None
            )
        except:
            identifier = "???"

        return "<{} {}={}>".format(
            self.__class__.__name__, self.primary_key(), identifier
        )

    def __hash__(self):
        return hash(self.identifier)

    def __eq__(self, other):
        return (
            hasattr(self, self.primary_key())
            and hasattr(other, self.primary_key())
            and self.identifier == other.identifier
        )

    def __setattr__(self, name, value):
        yield_callbacks = []
        attribute = self.attributes.get(name)

        if attribute and (
            attribute.indexes or attribute.cb_creation or attribute.cb_edition
        ):
            was_created = attribute.is_created(self)
            if not was_created:
                prev_value = None
                for callback in attribute.cb_creation:
                    res = callback(self, value)
                    if isinstance(res, types.GeneratorType):
                        try:
                            next(res)
                        except StopIteration:
                            pass
                        yield_callbacks.append(res)

            else:
                if any(index.primary for index in attribute.indexes.values()):
                    raise sheraf.SherafException(
                        f"Attribute {name} has a primary index and cannot be edited."
                    )

                prev_value = getattr(self, name)
                for callback in attribute.cb_edition:
                    res = callback(self, value, prev_value)
                    if isinstance(res, types.GeneratorType):
                        try:
                            next(res)
                        except StopIteration:
                            pass
                        yield_callbacks.append(res)

            old_values = self.before_index_edition(attribute)

        super().__setattr__(name, value)

        if attribute and (
            attribute.indexes or attribute.cb_creation or attribute.cb_edition
        ):
            try:
                self.after_index_edition(attribute, old_values)
            except sheraf.SherafException:
                if not was_created:
                    super().__delattr__(name)
                else:
                    super().__setattr__(name, prev_value)
                raise

            for callback in yield_callbacks:
                try:
                    next(callback)
                except StopIteration:
                    pass

    def __delattr__(self, name):
        yield_callbacks = []
        attribute = self.attributes.get(name)
        if attribute:
            old_values = self.before_index_edition(attribute)
            for callback in attribute.cb_deletion:
                res = callback(self, getattr(self, name))
                if isinstance(res, types.GeneratorType):
                    try:
                        next(res)
                    except StopIteration:
                        pass
                    yield_callbacks.append(res)

        super().__delattr__(name)

        if attribute:
            self.after_index_edition(attribute, old_values)
            for callback in yield_callbacks:
                try:
                    next(callback)
                except StopIteration:
                    pass

    def before_index_edition(self, attribute):
        old_index_values = {}
        for index in attribute.indexes.values():
            if not self._is_indexable(index):
                warnings.warn(
                    "New index in an already populated table. %s.%s will not be indexed. "
                    'Consider calling %s.index_table_rebuild(["%s"]) to initialize the indexation table.'
                    % (
                        self.__class__.__name__,
                        index.key,
                        self.__class__.__name__,
                        index.key,
                    ),
                    sheraf.exceptions.IndexationWarning,
                    stacklevel=5,
                )
                continue

            old_index_values[index] = index.get_model_values(self)
        return old_index_values

    def after_index_edition(self, attribute, old_index_values):
        for index in attribute.indexes.values():
            if not self._is_indexable(index):
                continue

            new_index_values = index.get_model_values(self)

            index_manager = self.indexes[index.key]
            index_manager.update_item(self, old_index_values[index], new_index_values)

    @property
    def identifier(self):
        """
        The identifier is the value of the primary_key for the current instance.
        If the primary_key is 'id', then the identifier might be an UUID.
        """
        if not self._identifier:
            self._identifier = getattr(self, self.primary_key())

        return self._identifier

    @property
    def raw_identifier(self):
        if not self._identifier:
            self._identifier = getattr(self, self.primary_key())

        if not self._raw_identifier:
            self._raw_identifier = self.mapping[self.primary_key()]

        return self._raw_identifier

    def _is_indexable(self, index):
        """
        To have its entries updated, an index must have its table previously
        initialized, with the exception of the very first model instance in
        the database.
        """
        if self._is_first_instance is None:
            self._is_first_instance = not self.index_manager().initialized()

        index_manager = self.indexes[index.key]
        index_table_exists = index_manager.table_initialized()
        return self._is_first_instance or index_table_exists

    def copy(self, **kwargs):
        r"""
        Copies a model.
        The attributes carrying an unique index wont be copied, they will be
        resetted instead.

        :param \*\*kwargs: Keywords arguments will be passed to
                         :func:`~sheraf.models.BaseModel.create` and thus
                         wont be copied.

        :return: a copy of this instance.
        """

        unique_attributes = (
            attribute
            for index in self.indexes.values()
            for attribute in index.details.attributes
            if index.details.unique
        )

        for attribute in unique_attributes:
            kwargs.setdefault(attribute.key(self), attribute.create(self))

        return super().copy(**kwargs)

    def delete(self):
        """Delete the current model instance.

        >>> class MyModel(sheraf.Model):
        ...     table = "my_model"
        ...
        >>> with sheraf.connection():
        ...    m = MyModel.create()
        ...    assert m == MyModel.read(m.id)
        ...    m.delete()
        ...    m.read(m.id)
        Traceback (most recent call last):
            ...
        sheraf.exceptions.ModelObjectNotFoundException: Id '...' not found in MyModel
        """
        for attr_name in self.attributes.keys():
            delattr(self, attr_name)

    @classmethod
    def count(cls, index_name=None):
        """
        Counts the number of elements in an index.
        :param index_name: The name of the index to count. By default
            the primary index is used
        """
        return cls.indexes[index_name or cls.primary_key()].count()


class IndexedModelMetaclass(BaseIndexedModelMetaclass):
    """
    Contains the mapping of tables (name of models) to their
    corresponding model definitions
    """

    tables = {}

    def __new__(cls, name, bases, attrs):
        klass = super().__new__(cls, name, bases, attrs)

        if "table" in attrs:
            table_name = attrs["table"]
            qualname = attrs["__module__"] + "." + attrs["__qualname__"]

            unique_table_name = attrs.get("unique_table_name", True)
            if (
                unique_table_name
                and table_name in IndexedModelMetaclass.tables
                and name != IndexedModelMetaclass.tables[table_name][0].split(".")[-1]
            ):
                message = "Table named '{table_name}' used twice: {first_class} and {second_class}".format(
                    table_name=table_name,
                    first_class=IndexedModelMetaclass.tables[table_name][0],
                    second_class=qualname,
                )
                raise sheraf.exceptions.SameNameForTableException(message)
            IndexedModelMetaclass.tables[table_name] = (qualname, klass)
        return klass


def model_from_table(table_name):
    if table_name not in IndexedModelMetaclass.tables:
        return None
    return IndexedModelMetaclass.tables[table_name][1]


class IndexedModel(BaseIndexedModel, metaclass=IndexedModelMetaclass):
    """
    :class:`~sheraf.models.indexation.IndexedModel` are the top-level
    models in the database. They come with one or several indexes,
    stored in a *table* at the root of the database. They must
    have a **table** parameter defined and an **id** attribute.

    They can have a **database_name** attribute. If it is set, then in a
    default connection context:

    - :func:`~sheraf.models.indexation.IndexedModel.create` will store the\
    new model instances in this database;
    - :func:`~sheraf.models.indexation.IndexedModel.read` and\
    :func:`~sheraf.models.indexation.IndexedModel.all` (etc.) will read in\
    priority in this database, and then in the default database.
    - :func:`~sheraf.models.indexation.IndexedModel.delete` will try to delete\
    the model from this database, and by default in the default database.

    However, if a **database_name** is explicitly passed to
    :func:`sheraf.databases.connection`, then every action will be
    performed on this database, ignoring the model **database_name** attribute.
    """

    database_name = None
    table = None

    id = sheraf.attributes.simples.SimpleAttribute().index(primary=True)

    @classmethod
    def index_manager(cls, index=None):
        return MultipleDatabaseIndexManager(
            cls.database_name,
            cls.table,
            index,
        )

    def __init__(self, *args, **kwargs):
        if "id" not in self.__class__.attributes:
            raise sheraf.exceptions.PrimaryKeyException(
                "{} inherit from IndexedModel but has no id attribute. Cannot create.".format(
                    self.__class__.__name__
                )
            )

        super().__init__(*args, **kwargs)


class SimpleIndexedModel(BaseIndexedModel, metaclass=BaseIndexedModelMetaclass):
    @classmethod
    def index_manager(cls, index=None):
        return SimpleIndexManager(index)



import click
import sheraf
import sys
from rich.traceback import install
from sheraf.health.utils import discover_models


install(show_locals=True)


@click.group()
@click.argument("uri")
def cli(uri):
    sys.path.append(".")
    db = sheraf.Database(uri)


@cli.command()
@click.argument("models", nargs=-1)
def check(models):
    with sheraf.connection():
        sheraf.print_health(*models)


@cli.command()
@click.argument("models")
@click.option(
    "--indexes",
    help="The name of the indexes to rebuild. If not provided all the indexes will be rebuilt.",
    default=None,
)
def rebuild(models, indexes):
    with sheraf.connection(commit=True):
        models = discover_models(models)
        for _, model in models:
            model.index_table_rebuild(indexes)



import contextlib
import os
import threading
import traceback

import transaction
import ZODB.config
import ZODB.DB
import zodburi
from ZODB.DemoStorage import DemoStorage

from sheraf.exceptions import ConnectionAlreadyOpened


class LocalData:
    instance = None

    def __init__(self, pid=None):
        self.pid = pid or os.getpid()
        self.databases = {}
        self.last_database_context = {}
        self.zodb_databases = {}

        class GlobalThreadContext(threading.local):
            def __init__(self):
                super().__init__()
                self.connections = []
                self.last_connection_context = None

        self.thread_context = GlobalThreadContext()

    @classmethod
    def get(cls):
        # TODO: put a lock on pid
        pid = os.getpid()
        if not cls.instance or cls.instance.pid != pid:
            cls.instance = LocalData(pid)
        return cls.instance


class Database(object):
    """A ZODB :class:`ZODB.DB` wrapper with a :class:`ZODB.interfaces.IStorage`
    factory.

    The storage factory will either create a
    :class:`~ZEO.ClientStorage.ClientStorage`, a
    :class:`~ZODB.FileStorage.FileStorage.FileStorage`, a
    :class:`~ZODB.DemoStorage.DemoStorage`, a Relstorage
    :class:`~relstorage.adapters.postgresql.adapter.PostgreSQLAdapter` or use a
    user defined storage depending on the argument passed at the initialization
    of the object.

    A Storage object is created and pass it to the :class:`ZODB.DB` constructor

    Several connections can be used at the same time. The connections are
    identified by their name.

    :param database_name: The name of the connection.
    :param storage: If set, this user defined storage will be used.
    :param uri: A zodburi to the database.
    :type uri: An URI that will be parsed by :func:`zodburi.resolve_uri`.
    :param db_args: Arguments to pass to the :class:`ZODB.DB`.

    :param nestable: If `False`, will raise a
        :class:`~sheraf.exceptions.ConnectionAlreadyOpened` if a connection has
        already been opened.
    """

    DEFAULT_DATABASE_NAME = "unnamed"

    def __init__(self, uri=None, storage=None, nestable=False, db_args=None):
        self.nestable = nestable
        self.uri = uri
        self.db = None
        self.storage = None
        self.db_args = db_args or {}

        class DatabaseThreadContext(threading.local):
            def __init__(self):
                super().__init__()
                self.connections = []

        self.thread_context = DatabaseThreadContext()

        self.db_args["databases"] = LocalData.get().zodb_databases

        self.reset(storage, uri)

        stack = traceback.extract_stack()[-2]
        LocalData.get().last_database_context[self.name] = (
            stack.filename,
            stack.lineno,
        )

    def __repr__(self):
        description = "<Database database_name='{}'".format(self.name)

        if self.db_args.get("read_only", False):
            description += " ro"

        if self.nestable:
            description += " nestable"

        description += ">"
        return description

    def reset(self, storage=None, uri=None):
        """Close and reopen a database connection."""
        if self.db:
            self.close()

        if storage is not None:
            self.storage = storage

        elif uri:
            storage_factory, db_args = zodburi.resolve_uri(uri)
            self.storage = storage_factory()
            db_args.update(self.db_args)
            self.db_args = db_args

        else:
            self.storage = DemoStorage()

        self.name = self.db_args.get("database_name", Database.DEFAULT_DATABASE_NAME)
        if self.name in LocalData.get().databases:
            last_context = LocalData.get().last_database_context.get(self.name)
            raise KeyError(
                "A database named '{}' already exists. Last opening was on {} at line {}".format(
                    self.name, last_context[0], last_context[1]
                )
                if last_context
                else "A database named '{}' already exists.".format(self.name)
            )

        self.db = ZODB.DB(self.storage, **self.db_args)
        LocalData.get().databases[self.name] = self

    def connection_open(self):
        """Opens a connection. Returns a connection to this database.

        If `nestable` is set and a connection has already been opened,
        raises a :class:`~sheraf.exceptions.ConnectionAlreadyOpened` exception.
        If `nestable` is False
        and a connection has already been opened, it returns a new connection
        with a new transaction_manager.

        :return: A :class:`~ZODB.Connection.Connection` object.
        """
        data = LocalData.get()

        # No other connection exists
        if not Database.last_connection():
            connection = self.db.open()

        # A connection to this database exists, and the second one is not allowed.
        elif not self.nestable:
            message = (
                "First connection was {} on {} at line {}".format(
                    Database.last_connection(),
                    *data.thread_context.last_connection_context
                )
                if data.thread_context.last_connection_context
                else "First connection was {}".format(Database.last_connection())
            )
            raise ConnectionAlreadyOpened(message)

        # A connection to this database exists, and the second one is allowed, but
        # with a new transaction manager.
        else:
            connection = self.db.open(
                transaction_manager=transaction.TransactionManager()
            )

        self.thread_context.connections.append(connection)
        data.thread_context.connections.append(connection)
        return connection

    def connection_close(self, connection=None):
        """Closes a connection opened on the database.

        :param connection: The connection to close, if `None` the last
            connection opened on the database is closed.
        """
        connection = connection or Database.last_connection(self)
        if connection.opened:
            connection.close()

        if connection in LocalData.get().thread_context.connections:
            LocalData.get().thread_context.connections.remove(connection)

        if connection in self.thread_context.connections:
            self.thread_context.connections.remove(connection)

    def close(self):
        """Closes the database."""
        data = LocalData.get()
        for connection in list(self.thread_context.connections):
            if connection and connection.opened:
                connection.close()
            data.thread_context.connections.remove(connection)
            self.thread_context.connections.remove(connection)

        if self.db:
            self.db.close()

        if self.name in data.databases:
            del data.databases[self.name]

        if self.name in data.zodb_databases:
            del data.zodb_databases[self.name]

        if self.name in data.last_database_context:
            del data.last_database_context[self.name]

        self.db = None
        self.storage = None

    @contextlib.contextmanager
    def connection(
        self, commit=False, cache_minimize=False, reuse=False, _trackeback_shift=0
    ):
        """A context manager opening a connection on this database.

        :param commit: Whether to commit the transaction when leaving the
            context manager.
        :type commit: boolean
        :param cache_minimize: Whether to call
            :func:`ZODB.Connection.Connection.cache_minimize` when leaving the
            context manager.
        :type cache_minimize: boolean
        :param reuse: If a connection is already opened, reuse it.
        :type reuse: boolean


        >>> database = sheraf.Database()
        >>> with database.connection() as connection:
        ...    sheraf.Database.current_connection() is connection
        True
        """
        if reuse and Database.last_connection(self):
            yield Database.last_connection(self)
            return

        _connection = self.connection_open()
        if not self.nestable:
            stack = traceback.extract_stack()[-3 - _trackeback_shift]
            LocalData.get().thread_context.last_connection_context = (
                stack.filename,
                stack.lineno,
            )

        try:
            yield _connection
            if commit:
                _connection.transaction_manager.commit()

        except BaseException:
            if commit and _connection.transaction_manager:
                _connection.transaction_manager.abort()
            raise

        finally:
            # TODO: to be changed with try/except NoTransaction when upgrading transaction>2.0 @cedric
            if not commit and _connection.transaction_manager:
                _connection.transaction_manager.abort()

            if cache_minimize:
                for conn in _connection.connections.values():
                    conn.cacheMinimize()

            self.connection_close(_connection)

    @classmethod
    def last_connection(cls, database=None):
        if database:
            return (
                database.thread_context.connections[-1]
                if database.thread_context.connections
                else None
            )
        return (
            LocalData.get().thread_context.connections[-1]
            if LocalData.get().thread_context.connections
            else None
        )

    @classmethod
    def current_connection(cls, database_name=None):
        if not Database.last_connection():
            return None

        if not database_name:
            return Database.last_connection()

        return Database.last_connection().get_connection(database_name)

    @classmethod
    def current_name(cls):
        if Database.last_connection():
            return Database.last_connection().db().database_name
        return None

    @classmethod
    def all(cls):
        """
        :return: A list containing all the existing :class:`Database` in a
            tuple `(name, Database)`.
        """
        return LocalData.get().databases.items()

    @classmethod
    def get(cls, database_name=None):
        """
        :param database_name: The name of the queried database.
        :return: The database object if it exists. A :class:`KeyError` is raised elsewise.
        """
        database_name = database_name or Database.DEFAULT_DATABASE_NAME
        try:
            return LocalData.get().databases[database_name]
        except KeyError:
            raise KeyError("No database named '{}'.".format(database_name))

    @classmethod
    def get_or_create(cls, **kwargs):
        """
        :return: The database object if it exists. If the database does not exist, it is created with the `kwargs` arguments.
        """
        try:
            return Database.get(database_name=kwargs.get("database_name"))
        except KeyError:
            return Database(**kwargs)


@contextlib.contextmanager
def connection(database_name=None, commit=False, cache_minimize=False, reuse=False):
    """
    Shortcut for :meth:`sheraf.databases.Database.connection`

    :param database_name: The name of the database on which to open a connection.
        If not set, the default database will be used.
    :param *kwargs: See :meth:`sheraf.databases.Database.connection` arguments.
    """
    database = Database.get(database_name)
    with database.connection(
        commit=commit,
        cache_minimize=cache_minimize,
        reuse=reuse,
        _trackeback_shift=2,
    ) as conn:
        yield conn



import importlib
import pkgutil

from sheraf.models.indexation import IndexedModelMetaclass


def import_submodules(*args):
    """Import all submodules of a module, recursively, including subpackages.

    :param *args: packages (name or actual module)
    :type *args: str | module
    :rtype: dict[str, types.ModuleType]
    """
    results = {}
    for package in args:
        if isinstance(package, str):
            try:
                # try as a module path
                package = importlib.import_module(package)
            except ModuleNotFoundError as exc:
                package = ".".join(package.split(".")[:-1])
                try:
                    # try as a module + model path
                    pkg = importlib.import_module(package)
                    return {package: pkg}

                except ModuleNotFoundError:
                    raise exc

        results[package.__name__] = package

        if not hasattr(package, "__path__"):
            continue

        for path, name, is_pkg in pkgutil.walk_packages(package.__path__):
            if not path.path.startswith(package.__path__[0]):
                continue

            full_name = package.__name__ + "." + name
            if is_pkg:
                results.update(import_submodules(full_name))
            else:
                results[full_name] = importlib.import_module(full_name)

    return results


def discover_models(*args):
    """Discover models in a module.

    :param args: Modules, or module strings.

    :return: A set of models.
    """

    modules = import_submodules(*args)
    result = {
        (path, model)
        for path, model in IndexedModelMetaclass.tables.values()
        if model.__module__ in modules.keys() or model in modules.keys()
    }

    for model in args:
        if isinstance(model, IndexedModelMetaclass):
            result.add(
                (
                    "{}.{}".format(model.__module__, model.__name__),
                    model,
                )
            )

    return result



import ZODB

import sheraf.attributes.models
import sheraf.models
import sheraf.models.inline
import sheraf.types


class Blob(sheraf.models.inline.InlineModel):
    """
    Blob objects wrap the raw data from regular files.

    ZODB :class:`ZODB.blob.Blob` are used to store the data.
    """

    @classmethod
    def create(cls, data=None, filename=None, stream=None, **kwargs):
        """
        :param data: the data to store in the blob
        :param filename: the name of the original file
        :param stream: If ``data`` is not set, data will be read from the ``stream`` stream.
        :param kwargs: optional attributes to set
        :return: A sheraf object wrapping a :class:`ZODB.blob.Blob` object, or ``None`` if this is an empty and unnamed file.
        """
        m = super().create(**kwargs)

        stream_data = stream.read() if stream else None
        if data or filename or stream_data:
            m.mapping["blob"] = ZODB.blob.Blob(data or stream_data)
            m.mapping["original_name"] = filename

        return m

    def open(self):
        """Opens the stored blob file."""
        return self.mapping["blob"].open()

    @property
    def data(self):
        """
        The file binary data.
        """
        f = self.mapping["blob"].open()
        try:
            return f.read()
        finally:
            f.close()

    @property
    def original_name(self):
        """
        The original filename.
        """
        return self.mapping["original_name"]

    @property
    def file_extension(self):
        """
        The original filename extension.
        """
        return self.original_name.split(".")[-1]

    @property
    def filename(self):
        """
        The name of the blob file.
        """
        f = self.mapping["blob"].open()
        try:
            return f.name
        finally:
            f.close()

    def delete(self):
        """
        Delete the object from the base.
        """
        self.mapping.clear()

    def __len__(self):
        return 1 if self.mapping else 0

    def __str__(self):
        return self.mapping.get("original_name")

    def __repr__(self):
        return '<Blob filename="{}">'.format(self.mapping.get("original_name"))

    def __bool__(self):
        return bool(self.mapping)

    def __getitem__(self, key):
        return getattr(self, key)

    def edit(
        self, value, addition=True, edition=True, deletion=False, replacement=False
    ):
        return value


class BlobAttribute(sheraf.attributes.models.InlineModelAttribute):
    """
    This attribute stores binary files.
    It is mainly an interface that handles a :class:`~sheraf.attributes.blobs.Blob` object.
    """

    def __init__(self, model=Blob, **kwargs):
        super().__init__(model=model, **kwargs)

    def serialize(self, value):
        if value is None:
            return None

        elif isinstance(value, sheraf.InlineModel):
            return value.mapping

        elif isinstance(value, dict):
            return self.model.create(**value).mapping

        elif hasattr(value, "filename") and (
            hasattr(value, "stream") or hasattr(value, "data")
        ):
            return Blob.create(
                stream=getattr(value, "stream", None),
                filename=getattr(value, "filename", None),
                data=getattr(value, "data", None),
            ).mapping

        else:
            return self._default_value(value)

    def deserialize(self, value):
        if not value:
            return None

        return self.model._decorate(value)



import datetime
import time

import sheraf.attributes
import sheraf.attributes.simples
from sheraf.models.base import BaseModel

DEPRECATED_META_DATETIME_FORMATS = ("%d/%m/%Y %H:%M:%S:%f", "%d/%m/%Y %H:%M")
DEPRECATED_DATE_TIME_FORMAT = "%d/%m/%Y %H:%M:%S:%f"


class IntAttributesModel(BaseModel):
    @classmethod
    def attribute_id(cls, name, attribute):
        return len(cls.attributes)


class NamedAttributesModel(BaseModel):
    @classmethod
    def attribute_id(cls, name, attribute):
        return name


class DatedNamedAttributesModel(NamedAttributesModel):
    """Model with creation and modification datetimes.

    Creation date is automatically saved. It will not change during
    object life. Date of modification is automatically saved when an
    attribute is modified and refers to the moment the transaction is
    committed. At creation time, date of modification and date of
    creation are equal.
    """

    _creation = sheraf.attributes.simples.SimpleAttribute(
        default=time.time,
        lazy=False,
    )

    def creation_datetime(self):
        """The date the object has been created. By now it refers to the date
        the method :func:`~sheraf.models.BaseModel.create` has been called, and
        not the date the transaction has been committed.

        :return: :class:`datetime.datetime` or None if the object has not been committed yet.
        """
        # TODO: The creation datetime should have the transaction commit datetime and not the object creation one

        if not self.mapping._p_oid:
            return None

        return self._deserialize_date(self._creation)

    def last_update_datetime(self):
        """The date of the last transaction commit involving a modification in
        this object.

        :return: :class:`datetime.datetime` or None if the object has not been committed yet.
        """
        return (
            self._deserialize_date(self.mapping._p_mtime)
            if self.mapping._p_mtime
            else None
        )

    @classmethod
    def _deserialize_date(cls, serialized_date):
        if isinstance(serialized_date, float):
            return datetime.datetime.fromtimestamp(serialized_date)

        for date_format in DEPRECATED_META_DATETIME_FORMATS:
            try:
                return datetime.datetime.strptime(serialized_date, date_format)
            except ValueError:
                pass

    def save(self):
        """Updates
        :func:`~sheraf.models.DatedNamedAttributesModel.last_update_datetime`
        value and saves all the model attributes."""

        self.mapping._p_changed = True
        return super().save()



import uuid
import random
import sys

from .attributes import (
    DatedNamedAttributesModel,
    IntAttributesModel,
    NamedAttributesModel,
)
from .indexation import SimpleIndexedModel, IndexedModel
from sheraf.attributes.simples import IntegerAttribute, StringUUIDAttribute


class UUIDIndexedModel:
    """Model using uuid4 as ids. Ids are stored as strings.

    >>> class MyUUIDModel(sheraf.IntIndexedModel):
    ...     table = "my_uuid_model"
    ...
    >>> with sheraf.connection():  # doctest: +SKIP
    ...     MyIntModel.create().id
    "e4bb714e-b5a8-40d6-bb69-ab3b932fbfe0"
    """

    id = StringUUIDAttribute(default=lambda: str(uuid.uuid4())).index(primary=True)


class IntIndexedModel:
    """Model using integers as ids.

    By default ids are 64bits integers.

    >>> class MyIntModel(sheraf.IntIndexedModel):
    ...     table = "my_int_model"
    ...
    >>> with sheraf.connection():  # doctest: +SKIP
    ...     MyIntModel.create().id
    383428472384721983
    """

    MAX_INT = sys.maxsize

    id = IntegerAttribute(default=lambda m: random.randint(0, m.MAX_INT)).index(
        primary=True
    )


class IntOrderedIndexedModel:
    """Model using increasing integers as ids.


    >>> class MyIntModel(sheraf.IntOrderedIndexedModel):
    ...     table = "my_int_model"
    ...
    >>> with sheraf.connection():  # doctest: +SKIP
    ...     MyIntModel.create().id
    ...     MyIntModel.create().id
    0
    1
    """

    id = IntegerAttribute(default=lambda m: m.count()).index(primary=True)


class IntIndexedNamedAttributesModel(
    NamedAttributesModel, IntIndexedModel, IndexedModel
):
    """The ids of this model are integers, and attributes are named."""


class IntOrderedNamedAttributesModel(
    NamedAttributesModel, IntOrderedIndexedModel, IndexedModel
):
    """The ids are 64bits integers, distributed ascendently starting at 0."""


class UUIDIndexedNamedAttributesModel(
    NamedAttributesModel, UUIDIndexedModel, IndexedModel
):
    """The ids of this model are UUID4, and attributes are named."""


class UUIDIndexedDatedNamedAttributesModel(
    DatedNamedAttributesModel, UUIDIndexedModel, IndexedModel
):
    """The ids of this model are UUID4, the attributes are named, and any
    modification on the model will update its modification datetime."""


class IntIndexedIntAttributesModel(IntAttributesModel, IntIndexedModel, IndexedModel):
    """The ids of this models are integers, and the ids of its attributes are
    also integers."""


class AttributeModel(NamedAttributesModel, SimpleIndexedModel):
    """
    This model is expected to be used with :class:`~sheraf.attributes.models.IndexedModelAttribute`.
    Its usage is mainly the same as any :class:`~sheraf.models.indexation.BaseIndexedModel`.
    """


Model = UUIDIndexedDatedNamedAttributesModel



from rich.console import Console
from rich.progress import Progress, BarColumn, TextColumn, TimeRemainingColumn
from rich.table import Table
from rich.text import Text
from rich.markdown import Markdown
from sheraf.health.utils import discover_models

from .checks import check_conflict_resolution, check_attributes_index, check_model_index


# Set the list of check functions to be run
INSTANCE_CHECK_FUNCS = {}
ATTRIBUTE_CHECK_FUNCS = {
    "index": check_attributes_index,
}
MODEL_CHECK_FUNCS = {"index": check_model_index}
OTHER_CHECK_FUNCS = {"check_conflict_resolution": check_conflict_resolution}


def check_health(
    *args,
    model_checks=None,
    instance_checks=None,
    attribute_checks=None,
    other_checks=None,
    console=None,
):
    """
    Takes some modules in parameters.

    :param model_checks: If None, check also model consistency (see constant model_check_funcs)
    :param instance_checks: If None, check all instance consistency rules. Else, give the list of wanted
        checking rules (see constant instance_check_funcs)
    :param attribute_checks: If None, check all attribute consistency rules. Else, give the list of wanted
        checking rules (see constant instance_check_funcs)
        The function will discover models in the modules, analyze every
        model instance, and return a health report in JSON. Depending on the
        result, you may need to apply some fixes.
    """
    models = discover_models(*args)
    health_report = {}

    instance_checks = instance_checks or []
    attribute_checks = attribute_checks or []
    model_checks = model_checks or []
    other_checks = other_checks or []
    if not instance_checks and not attribute_checks and not model_checks:
        instance_checks = INSTANCE_CHECK_FUNCS.keys()
        attribute_checks = ATTRIBUTE_CHECK_FUNCS.keys()
        model_checks = MODEL_CHECK_FUNCS.keys()
        other_checks = OTHER_CHECK_FUNCS.keys()

    for check_key in other_checks:
        health_report[check_key] = OTHER_CHECK_FUNCS[check_key]()

    with Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(bar_width=None),
        TextColumn("{task.completed}", justify="right"),
        TextColumn("/"),
        TextColumn("{task.total}"),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        TimeRemainingColumn(),
        console=console,
        transient=True,
    ) as progress:
        tasks = {}
        for model_path, model in models:
            tasks[model] = progress.add_task(model.__name__)

        for model_path, model in models:
            progress.update(tasks[model], total=model.count())

        for model_path, model in models:
            progress.start_task(tasks[model])

            for check_key in instance_checks:
                instance_check_func = INSTANCE_CHECK_FUNCS[check_key]
                health_report.setdefault(instance_check_func.__name__, {}).setdefault(
                    model_path, {"ok": 0, "ko": 0}
                )

            for check_key in model_checks:
                model_check_func = MODEL_CHECK_FUNCS[check_key]
                health_report.setdefault(model_check_func.__name__, {})[
                    model_path
                ] = model_check_func(model)

            # Iterate on instances
            for m in model.all():
                progress.update(tasks[model], advance=1)
                for check_key in instance_checks:
                    check_func = INSTANCE_CHECK_FUNCS[check_key]
                    model_instance_result = health_report[check_func.__name__][
                        model_path
                    ]
                    if check_func(m):
                        model_instance_result["ok"] += 1
                    else:
                        model_instance_result["ko"] += 1

                for check_key in attribute_checks:
                    check_func = ATTRIBUTE_CHECK_FUNCS[check_key]
                    health_report.setdefault(check_func.__name__, {})
                    for attribute_name, bool_value in check_func(m).items():
                        attribute_result = (
                            health_report[check_func.__name__]
                            .setdefault(model_path, {})
                            .setdefault(attribute_name, {"ok": 0, "ko": 0})
                        )
                        if bool_value:
                            attribute_result["ok"] += 1
                        else:
                            attribute_result["ko"] += 1
    return health_report


def _print_check_other_health_result(console, check_reason, health_table, help):
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title=check_reason,
        expand=True,
        caption=help,
    )
    table.add_column("Check")
    table.add_column("State")

    table.add_row(check_reason, "OK" if health_table[check_reason] else "KO")
    console.print(table)


def _print_check_model_health_result(console, check_reason, health_table, help):
    table_key = "check_model_" + check_reason
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title=table_key,
        expand=True,
        caption=help,
    )
    table.add_column("Model")
    table.add_column("KO")
    table.add_column("OK")

    for model_path, attributes in health_table.get(table_key, {}).items():
        table.add_row(
            model_path,
            str(sum(values["ko"] for values in attributes.values())),
            str(sum(values["ok"] for values in attributes.values())),
        )

        for attribute_name, values in attributes.items():
            table.add_row(f"  - {attribute_name}", str(values["ko"]), str(values["ok"]))

    if health_table.get(table_key):
        console.print(table)
    else:
        console.print("  No model to visit.")


def _print_check_instance_health_result(console, check_reason, health_table, help):
    """
    :param check_reason: one among model_checks keys
    :param health_table: result of a check function
    """
    table_key = "check_instance_" + check_reason
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title=table_key,
        expand=True,
        caption=help,
    )
    table.add_column("Model")
    table.add_column("KO")
    table.add_column("OK")
    for model_path, values in health_table.get(table_key, {}).items():
        table.add_row(model_path, str(values["ko"]), str(values["ok"]))
    console.print(table)


def _print_check_attribute_health_result(console, check_reason, health_table, help):
    table_key = "check_attributes_" + check_reason
    table = Table(
        show_header=True,
        header_style="bold magenta",
        title=table_key,
        expand=True,
        caption=help,
    )
    table.add_column("Model")
    table.add_column("KO")
    table.add_column("OK")

    for model_path, attributes in health_table.get(table_key, {}).items():
        table.add_row(
            model_path,
            str(sum(values["ko"] for values in attributes.values())),
            str(sum(values["ko"] for values in attributes.values())),
        )

        for attribute_name, values in attributes.items():
            table.add_row(f"  - {attribute_name}", str(values["ko"]), str(values["ok"]))

    if health_table.get(table_key):
        console.print(table)
    else:
        console.print("  No model to visit.")


def print_health(
    *args,
    model_checks=None,
    instance_checks=None,
    attribute_checks=None,
    other_checks=None,
):
    console = Console(width=100)
    """Takes some modules in parameters (e.g. "american._class.cowboy_module").

    The function will discover models in the modules, analyze every model instance, and return
    a health report in a human readable format. Depending on the result, you may need to apply some fixes.

    This function does not edit any data and is safe to be executed in a production shell.
    """
    console.print(
        Text(
            "====== _ ==================  __ ==\n"
            "=     | |                   / _| =\n"
            "=  ___| |__   ___ _ __ __ _| |_  =\n"
            "= / __| '_ \\ / _ \\ '__/ _` |  _| =\n"
            "= \\__ \\ | | |  __/ | | (_| | |   =\n"
            "= |___/_| |_|\\___|_|  \\__,_|_|   =\n"
            "==================================",
            justify="center",
        )
    )

    instance_checks = instance_checks or []
    attribute_checks = attribute_checks or []
    model_checks = model_checks or []
    other_checks = other_checks or []
    if not instance_checks and not attribute_checks and not model_checks:
        instance_checks = INSTANCE_CHECK_FUNCS.keys()
        attribute_checks = ATTRIBUTE_CHECK_FUNCS.keys()
        model_checks = MODEL_CHECK_FUNCS.keys()
        other_checks = OTHER_CHECK_FUNCS.keys()

    health = check_health(
        *args,
        model_checks=model_checks,
        instance_checks=instance_checks,
        attribute_checks=attribute_checks,
        other_checks=other_checks,
        console=console,
    )

    for other_check_type in other_checks:
        console.print()
        _print_check_other_health_result(
            console,
            other_check_type,
            health,
            Markdown(OTHER_CHECK_FUNCS[other_check_type].__doc__),
        )
        console.print()

    for model_check_type in model_checks:
        console.print()
        _print_check_model_health_result(
            console,
            model_check_type,
            health,
            Markdown(MODEL_CHECK_FUNCS[model_check_type].__doc__),
        )
        console.print()

    for instance_check_type in instance_checks:
        console.print()
        _print_check_instance_health_result(
            console,
            instance_check_type,
            health,
            Markdown(INSTANCE_CHECK_FUNCS[instance_check_type].__doc__),
        )
        console.print()

    for attribute_check_type in attribute_checks:
        console.print()
        _print_check_attribute_health_result(
            console,
            attribute_check_type,
            health,
            Markdown(ATTRIBUTE_CHECK_FUNCS[attribute_check_type].__doc__),
        )
        console.print()



import ZODB

import sheraf


def check_conflict_resolution():
    """
    Checks wether sheraf object conflicts resolutions are possible.
    When this is KO, it is generally because sheraf is not installed in the ZEO
    environnement, so ZEO cannot solve sheraf object conflicts.
    """
    table_name = "__conflict_resolution_test_model__"
    nestable = sheraf.Database.get().nestable
    sheraf.Database.get().nestable = True

    class TestModel(sheraf.Model):
        table = table_name
        counter = sheraf.CounterAttribute(default=0)

    with sheraf.connection(commit=True):
        m = TestModel.create()

    try:
        with sheraf.connection(commit=True):
            TestModel.read(m.identifier).counter.increment(1)

            with sheraf.connection(commit=True):
                TestModel.read(m.identifier).counter.increment(15)

    except ZODB.POSException.ConflictError:  # pragma: no cover
        return False

    else:
        return True

    finally:
        with sheraf.connection(commit=True) as conn:
            del conn.root()[table_name]

        sheraf.Database.get().nestable = nestable


def check_attributes_index(model_instance):
    """
    Given a model instance, computes all the values for all the indexes,
    then checks the index table if the values match the model instance.

    This finds instances that are not synchronized with theirs indexes.
    """
    root = sheraf.Database.current_connection().root()
    result = {}
    index_table = root.get(model_instance.table)

    if not index_table:
        return result

    for index_name, index in model_instance.indexes.items():
        if index.details.primary:
            continue

        values = index.details.get_model_values(model_instance)

        if values and index_name not in index_table:
            result[index_name] = False
            continue

        if index.details.unique:
            result[index_name] = all(
                value in index_table[index_name]
                and index_table[index_name][value] == model_instance.mapping
                for value in values
            )
        else:
            result[index_name] = all(
                value in index_table[index_name]
                and model_instance.mapping
                == index_table[index_name][value][model_instance.raw_identifier]
                for value in values
            )
    return result


def check_model_index(model):
    """
    Given a model, for every index table, reads every mapping in the table
    and checks that the mappin belongs a model.

    This finds orphans models.
    """
    root = sheraf.Database.current_connection().root()
    index_table = root.get(model.table)
    result = {}

    if not index_table:
        return result

    for attribute_index_key, attribute_index_table in index_table.items():
        index = model.indexes[attribute_index_key]

        if index.details.primary:
            continue

        for mmapping in attribute_index_table.values():
            try:
                if index.details.unique:
                    model.read(model._decorate(mmapping).identifier)
                else:
                    [
                        model.read(model._decorate(persistent).identifier)
                        for persistent in mmapping.values()
                    ]
                result.setdefault(attribute_index_key, {"ok": 0, "ko": 0})["ok"] += 1

            except sheraf.exceptions.ModelObjectNotFoundException:
                result.setdefault(attribute_index_key, {"ok": 0, "ko": 0})["ko"] += 1

    return result



from BTrees.OOBTree import OOBTree


class Index:
    """
    Indexes should be either created as :class:`~sheraf.models.indexation.IndexedModel` class parameters,
    or with the attributes :func:`~sheraf.attributes.Attribute.index` method.

    :param attributes: The attributes being indexed. They can be either
        :class:`~sheraf.attributes.Attribute` or strings representing
        the attributes names in the model.
    :type *attributes: A collection of :class:`~sheraf.attributes.Attribute` or strings.
    :param key: The key the index will use. By default, it takes the name it has
                as a :class:`~sheraf.models.indexation.IndexedModel` attribute.
                If the :func:`~sheraf.attributes.Attribute.index` is used,
                the key is the :class:`~sheraf.attributes.Attribute` name.
    :param unique: If the index is unique, and two models have the same value for this
                   model, a :class:`~sheraf.exceptions.UniqueIndexException` is raised
                   when trying to write the second one. Automatically set to :class:`True` if
                   primary is :class:`True`.
    :type unique: bool
    :param values: A callable that takes the current attribute value and returns a
                   collection of values to index. Each generated value will be
                   indexed each time this attribute is edited. It may take time if
                   the generated collection is large. By default the attribute
                   :meth:`~sheraf.attributes.Attribute.values` method is
                   applied.
    :param search: A callable that takes some raw data and returns a collection
                   of values to search in the index. By default, the
                   :meth:`~sheraf.attributes.Attribute.search` method is
                   used.
    :param mapping: The mapping object to be used to store the indexed values.
                   By default :class:`~BTrees.OOBTree.OOBTree` is used.
    :param nullok: If `True`, `None` or empty values can be indexed. `True` by default.
    :param noneok: Ignored in if `nullok` is `True`. Else, if `noneok` is
                   `True`, `None` values can be indexed. `False` by default."

    >>> class People(sheraf.Model):
    ...     table = "index_people"
    ...
    ...     # Simple indexing
    ...     name = sheraf.SimpleAttribute()
    ...     nameindex = sheraf.Index("name")
    ...
    ...     # Indexing with the .index() shortcut
    ...     size = sheraf.IntegerAttribute().index()
    ...
    ...     # Emails can only be owned once
    ...     email = sheraf.SimpleAttribute().index(unique=True)
    ...
    ...     # Indexing people by their decade
    ...     age = sheraf.SimpleAttribute().index(key="decade", values=lambda age: {age // 10})
    ...
    >>> with sheraf.connection(commit=True):
    ...     m = People.create(
    ...         name="George Abitbol",
    ...         size=180,
    ...         email="george@abitbol.com",
    ...         age=55,
    ...     )
    ...
    >>> with sheraf.connection():
    ...     assert [m] == People.filter(nameindex="George Abitbol")
    ...     assert [m] == People.filter(size=180)
    ...     assert [m] == People.filter(decade=5)
    ...
    >>> with sheraf.connection():
    ...     People.create(name="Peter", size=175, email="george@abitbol.com", age=35)
    Traceback (most recent call last):
        ...
    UniqueIndexException
    """

    def __init__(
        self,
        *attributes,
        unique=False,
        key=None,
        values=None,
        search=None,
        mapping=None,
        primary=False,
        nullok=None,
        noneok=None,
    ):
        self.attributes = attributes
        self.unique = unique or primary
        self.key = key
        self.default_values_func = values
        self.search_func = search or values
        self.values_funcs = {}
        self.mapping = mapping or OOBTree
        self.primary = primary
        self.nullok = nullok
        self.noneok = noneok

    def __repr__(self):
        if self.primary:
            return "<Index key={} unique={} primary>".format(self.key, self.unique)
        return "<Index key={} unique={}>".format(self.key, self.unique)

    def values(self, *args, **kwargs):
        """
        This decorator sets a values method, for one, several or all of the index attributes.

        - If no positionnal argument is passed, then this sets a default values function for the
          index.
        - You can pass one or several attributes or attributes names to set a specific values method
          for those attributes.
        - If you do set a specific values method for one or several attributes, the decorated function
          will be given the attribute values as positionnal arguments, in the order they were passed
          to the decorator.

        The decorated function must return a single key, or a collection of keys, where the index
        will store the current model instance. Depending on the `noneok` and `nullok`
        :class:`~sheraf.attributes.index.Index` parameters, `None` and falsy index keys might be
        ignored.

        >>> class Cowboy(sheraf.Model):
        ...     table = "any_values_table"
        ...     first_name = sheraf.StringAttribute()
        ...     last_name = sheraf.StringAttribute()
        ...     surname = sheraf.StringAttribute()
        ...     theindex = sheraf.Index(first_name, last_name, surname)
        ...
        ...     @theindex.values
        ...     def default_name_indexation(self, values):
        ...         return values.lower()
        ...
        ...     @theindex.values(first_name, "last_name")
        ...     def full_name(self, first, last):
        ...         return f"{first} {last}".lower()
        ...
        >>> with sheraf.connection():
        ...     m = Cowboy.create(first_name="George", last_name="Abitbol", surname="Georgy")
        ...     assert m in Cowboy.filter(theindex="george abitbol")
        ...     assert m in Cowboy.filter(theindex="georgy")
        """

        # Guess if the decorator has been called with or without parenthesis
        if args and (len(args) > 1 or not callable(args[0])):
            attributes = args
            args = []
        else:
            attributes = []

        def wrapper(func):
            # If no attribute is passed as a positionnal argument,
            # the method will be the default values method
            if not attributes:
                self.default_values_func = func

                if self.search_func is None:
                    self.search_func = func

            # Else the method will be assigned to each attribute
            self.values_funcs.setdefault(func, []).append(attributes)

            return func

        return wrapper if not args else wrapper(args[0])

    def search(self, *args, **kwargs):
        """
        This decorator sets the search method for the index. It should return
        a single key to search in the index, or a collection of keys to
        search in the index. If it returns an indexed collection, by default the
        model instances will be returned in the order the index keys will be iterated.

        >>> class Model(sheraf.Model):
        ...     table = "any_search_table"
        ...     foo = sheraf.StringAttribute()
        ...     bar = sheraf.StringAttribute()
        ...     theindex = sheraf.Index(foo, bar)
        ...
        ...     @theindex.search
        ...     def the_search_method(self, values):
        ...         return values.lower()
        ...
        >>> with sheraf.connection():
        ...     m = Model.create(foo="foo", bar="bar")
        ...     assert m in Model.search(theindex="foo")
        ...     assert m in Model.search(theindex="BAR")
        """

        def wrapper(func):
            self.search_func = func
            return func

        return wrapper if not args else wrapper(args[0])

    def get_model_values(self, model):
        return {
            v
            for func, attr_groups in self.values_funcs.items()
            for attributes in attr_groups
            for v in self.get_values(model, attributes, func)
        }

    def get_values(self, model, attributes, func):
        values = self.call_values_func(model, attributes, func)

        if not self.nullok:  # Empty values are not indexed
            return {v for v in values if v}

        elif not self.noneok:  # None values are not indexed
            return {v for v in values if v is not None}

        else:  # Everything is indexed
            return values

    def call_values_func(self, model, attributes, func):
        if not all(attribute.is_created(model) for attribute in attributes):
            return {}

        values = [attribute.read(model) for attribute in attributes]

        if not func:
            return set(values)

        try:
            values = func(*values)
        except TypeError as exc:
            if "positional argument" not in str(exc):
                raise
            values = func(model, *values)

        values = values if isinstance(values, (list, set, tuple, dict)) else {values}
        return values

    def call_search_func(self, model, value):
        if not self.search_func:
            return {value}

        try:
            values = self.search_func(value)
        except TypeError as exc:
            if "positional argument" not in str(exc):
                raise
            values = self.search_func(model, value)

        values = values if isinstance(values, (list, set, tuple, dict)) else {values}
        return values
